# Query responses
EXPECTED_GENERATE_TEMPLATE_GQL_RESPONSE_FIELD = 'generateCollectorTemplate'
EXPECTED_TTC_RESPONSE_FIELD = 'testTelnetConnection'
EXPECTED_TTOC_RESPONSE_FIELD = 'testTcpOpenConnection'
EXPECTED_ADD_DC_RESPONSE_FIELD = 'createCollectorRecord'

DEFAULT_COLLECTION_REGION = 'us-east-1'

# Verbiage
ADD_DC_REGION_PROMPT_VERBIAGE = 'Which AWS region would you like to deploy in?'
ADD_DC_PROMPT_VERBIAGE = 'Do you want deploy this collector right now? This opens a browser to CF console with a quick create link'


import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.integrations.onboarding.bi.reports import ReportsOnboardingService
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.integrations.info.status import OnboardingStatusService
from montecarlodata.integrations.onboarding.data_lake.events import EventsOnboardingService
from montecarlodata.integrations.onboarding.data_lake.glue_athena import GlueAthenaOnboardingService
from montecarlodata.integrations.onboarding.data_lake.hive import HiveOnboardingService
from montecarlodata.integrations.onboarding.data_lake.presto import PrestoOnboardingService
from montecarlodata.integrations.onboarding.data_lake.spark import SparkOnboardingService, \
    SPARK_HTTP_MODE_CONFIG_TYPE, SPARK_DATABRICKS_CONFIG_TYPE, SPARK_BINARY_MODE_CONFIG_TYPE
from montecarlodata.integrations.onboarding.fields import (
    CONNECTION_TO_WAREHOUSE_TYPE_MAP,
    GLUE_CONNECTION_TYPE,
    GQL_TO_FRIENDLY_CONNECTION_MAP,
    HIVE_MYSQL_CONNECTION_TYPE,
    HIVE_S3_CONNECTION_TYPE,
    PRESTO_S3_CONNECTION_TYPE,
    S3_METADATA_EVENT_TYPE,
    S3_QL_EVENT_TYPE,
    SECRETS_MANAGER_CREDENTIAL_MECHANISM,
    SELF_HOSTING_MECHANISMS,
)
from montecarlodata.integrations.onboarding.operations.connection_ops import ConnectionOperationsService
from montecarlodata.integrations.onboarding.self_hosted_credentials import SelfHostedCredentialOnboardingService
from montecarlodata.integrations.onboarding.warehouse.warehouses import WarehouseOnboardingService
from montecarlodata.integrations.keys import IntegrationKeyScope, IntegrationKeyService
from montecarlodata.tools import add_common_options, validate_json_callback, AdvancedOptions, \
    convert_uuid_callback, convert_empty_str_callback

# Shared command verbiage
METADATA_VERBIAGE = 'For metadata'
QL_VERBIAGE = 'For query logs'
SQL_VERBIAGE = 'For health queries'
EVENT_VERBIAGE = 'For tracking data freshness and volume at scale. Requires s3 notifications to be configured first'
REGION_VERBIAGE = 'If not specified the region the collector is deployed in is used'
WAREHOUSE_VERBIAGE = 'For metadata, query logs and metrics'
BI_VERBIAGE = 'For reports'
PASSWORD_VERBIAGE = f'If you prefer a prompt (with hidden input) enter {settings.SHOW_PROMPT_VALUE}'
RESOURCE_VERBIAGE = 'This can be helpful if the resources are in different accounts'

# Options shared across commands
ROLE_OPTIONS = [
    click.option('--role', help='Assumable role ARN to use for accessing AWS resources.', required=False),
    click.option('--external-id', help='An external id, per assumable role conditions.', required=False),
]
S3_OPTIONS = [
    click.option('--bucket', help='S3 Bucket where query logs are contained.', required=True),
    click.option('--prefix', help='Path to query logs.', required=True),
    *ROLE_OPTIONS
]
EVENT_OPTIONS = [
    click.option('--enable/--disable', 'toggle', help='Enable or disable events. Enables if not specified.',
                 default=True)
]
DATABASE_OPTIONS = [
    click.option('--host', help='Hostname.', required=True),
    click.option('--user', help='Username with access to the database.', required=True),
    click.option('--database', help='Name of database.', required=True),
    click.option('--password', help=f'User\'s password. {PASSWORD_VERBIAGE}.', required=True, cls=AdvancedOptions, prompt_if_requested=True)
]

CONNECTION_OPTIONS = [
    click.option('--connection-id', help='ID for the connection.', required=True, type=click.UUID, callback=convert_uuid_callback)
]

ONBOARDING_CONFIGURATION_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option('--skip-validation', is_flag=True, help='Skip all connection tests.',
                 cls=AdvancedOptions, mutually_exclusive_options=['validate_only']),
    click.option('--validate-only', is_flag=True, help='Run connection tests without adding.',
                 cls=AdvancedOptions, mutually_exclusive_options=['skip_validation']),
    click.option('--auto-yes', is_flag=True, help='Skip any interactive approval.', default=False, show_default=True)
]

NETWORK_OPTIONS = [
    click.option('--test-network-only', 'skip_permission_tests', is_flag=True, default=False, show_default=True,
                 cls=AdvancedOptions, mutually_exclusive_options=['skip_validation'],
                 help='Skip any permission tests. Only validates network connection between the collector and resource can be established.')
]

BI_OPTIONS = [
    click.option('--verify-ssl/--no-verify-ssl', 'verify_ssl', required=False, default=True, show_default=True,
                 help='Whether to verify the SSL connection (uncheck for self-signed certs).')
]


@click.group(help='Manage or integrate an asset with Monte Carlo.')
def integrations():
    """
    Group for any integration related subcommands
    """
    pass


@integrations.command(help=f'Setup a Hive metastore integration (MySQL). {METADATA_VERBIAGE}.')
@click.pass_obj
@click.option('--port', help='HTTP port.', default=3306, type=click.INT, show_default=True)
@click.option('--use-ssl', help='Use SSL to connect (using AWS RDS certificates).', required=False, is_flag=True,
              default=False, show_default=True)
@click.option('--catalog', help='Presto catalog name. For using multiple hive clusters with Presto. '
                                'Uses \'hive\' if not specified.', required=False)
@add_common_options(DATABASE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive_metastore(ctx, database, **kwargs):
    """
    Onboard a hive metastore connection (MySQL)
    """
    HiveOnboardingService(config=ctx['config']).onboard_hive_mysql(dbName=database, **kwargs)


@integrations.command(help=f'Setup a Presto SQL integration. {SQL_VERBIAGE}.')
@click.pass_obj
@click.option('--host', help='Hostname.', required=True)
@click.option('--port', help='HTTP port.', default=8889, type=click.INT, show_default=True)
@click.option('--user', help='Username with access to catalog/schema.', required=False)
@click.option('--password', help=f'User\'s password. {PASSWORD_VERBIAGE}', required=False, cls=AdvancedOptions, prompt_if_requested=True)
@click.option('--catalog', help='Mount point to access data source.', required=False)
@click.option('--schema', help='Schema to access.', required=False)
@click.option('--http-scheme', help='Scheme for authentication.',
              type=click.Choice(['http', 'https'], case_sensitive=True), required=True)
@click.option('--cert-file', help='Local SSL certificate file to upload to collector.', required=False,
              type=click.Path(dir_okay=False, exists=True), cls=AdvancedOptions, mutually_exclusive_options=['cert_s3'])
@click.option('--cert-s3', help='Object path (key) to a certificate already uploaded to the collector.',
              required=False, cls=AdvancedOptions, mutually_exclusive_options=['cert_file'])
@click.option('--skip-cert-verification', help='Skip SSL certificate verification.', required=False, is_flag=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_presto(ctx, password, **kwargs):
    """
    Onboard a presto sql connection
    """
    if not password:
        password = None  # make explicitly null if not set. Prompts can't be None
    PrestoOnboardingService(config=ctx['config']).onboard_presto_sql(password=password, **kwargs)


@integrations.command(help=f'Setup a Hive SQL integration. {SQL_VERBIAGE}.')
@click.pass_obj
@click.option('--host', help='Hostname.', required=True)
@click.option('--database', help='Name of database.', required=False)
@click.option('--port', help='HTTP port.', default=10000, type=click.INT, show_default=True)
@click.option('--user', help='Username with access to hive.', required=True)
@click.option('--auth-mode', help='Hive authentication mode.', required=False, default='SASL',
              type=click.Choice(['SASL', 'NOSASL']), show_default=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive(ctx, user, **kwargs):
    HiveOnboardingService(config=ctx['config']).onboard_hive_sql(username=user, **kwargs)


@integrations.command(help=f'Setup a Presto logs integration (S3). {QL_VERBIAGE}.')
@click.pass_obj
@add_common_options(S3_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_presto_logs(ctx, role, **kwargs):
    """
    Onboard a presto logs (s3) connection
    """
    PrestoOnboardingService(config=ctx['config']).onboard_presto_s3(assumable_role=role, **kwargs)


@integrations.command(help=f'Setup a Hive EMR logs integration (S3). {QL_VERBIAGE}.')
@click.pass_obj
@add_common_options(S3_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive_logs(ctx, role, **kwargs):
    """
    Onboard a hive emr (s3) connection
    """
    HiveOnboardingService(config=ctx['config']).onboard_hive_s3(assumable_role=role, **kwargs)


@integrations.command(help=f'Setup a Glue integration. {METADATA_VERBIAGE}.')
@click.pass_obj
@click.option('--region', help=f'Glue catalog region. {REGION_VERBIAGE}.', required=False)
@add_common_options(ROLE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_glue(ctx, role, region, **kwargs):
    """
    Onboard a glue connection
    """
    GlueAthenaOnboardingService(config=ctx['config']).onboard_glue(assumable_role=role, aws_region=region, **kwargs)


@integrations.command(help='Setup an Athena integration. For query logs and health queries.')
@click.pass_obj
@click.option('--catalog', help='Glue data catalog. If not specified the AwsDataCatalog is used.', required=False)
@click.option('--workgroup',
              help='Workbook for running queries and retrieving logs. If not specified the primary is used.',
              required=False)
@click.option('--region', help=f'Athena cluster region. {REGION_VERBIAGE}.', required=False)
@add_common_options(ROLE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_athena(ctx, role, region, **kwargs):
    """
    Onboard an athena connection
    """
    GlueAthenaOnboardingService(config=ctx['config']).onboard_athena(assumable_role=role, aws_region=region, **kwargs)


@integrations.command(help=f'Setup a thrift binary Spark integration. {SQL_VERBIAGE}.')
@click.pass_obj
@click.option('--host', help='Hostname.', required=True)
@click.option('--database', help='Name of database.', required=True)
@click.option('--port', help='Port.', default=10000, type=click.INT, show_default=True)
@click.option('--user', help='Username with access to spark.', required=True)
@click.option('--password', help=f'User\'s password. {PASSWORD_VERBIAGE}.', required=True,
              cls=AdvancedOptions, prompt_if_requested=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_binary_mode(ctx, user, **kwargs):
    """
    Onboard a spark connection, thrift binary mode
    """
    SparkOnboardingService(config=ctx['config']).onboard_spark(SPARK_BINARY_MODE_CONFIG_TYPE, username=user, **kwargs)


@integrations.command(help=f'Setup a thrift HTTP Spark integration. {SQL_VERBIAGE}.')
@click.pass_obj
@click.option('--url', help='HTTP URL.', required=True)
@click.option('--user', help='Username with access to spark.', required=True)
@click.option('--password', help=f'User\'s password. {PASSWORD_VERBIAGE}.', required=True,
              cls=AdvancedOptions, prompt_if_requested=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_http_mode(ctx, user, **kwargs):
    """
    Onboard a spark connection, thrift http mode
    """
    SparkOnboardingService(config=ctx['config']).onboard_spark(SPARK_HTTP_MODE_CONFIG_TYPE, username=user, **kwargs)


@integrations.command(help=f'Setup a Spark integration for Databricks. {SQL_VERBIAGE}.')
@click.pass_obj
@click.option('--workspace-url', help='Databricks workspace URL.', required=True)
@click.option('--workspace-id', help='Databricks workspace ID.', required=True)
@click.option('--cluster-id', help='Databricks cluster ID.', required=True)
@click.option('--token', help=f'Databricks access token. {PASSWORD_VERBIAGE}.', required=True,
              cls=AdvancedOptions, prompt_if_requested=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_databricks(ctx, **kwargs):
    """
    Onboard a spark connection, databricks
    """
    SparkOnboardingService(config=ctx['config']).onboard_spark(SPARK_DATABRICKS_CONFIG_TYPE, **kwargs)


@integrations.command(help=f'Setup a Redshift integration. {WAREHOUSE_VERBIAGE}.')
@click.pass_obj
@click.option('--port', help='HTTP port.', default=5439, type=click.INT, show_default=True)
@click.option('--name', help='Friendly name for the warehouse.', required=True)
@add_common_options(DATABASE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@add_common_options(NETWORK_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_redshift(ctx, database, name, **kwargs):
    """
    Onboard a redshift connection
    """
    WarehouseOnboardingService(config=ctx['config']).onboard_redshift(dbName=database, warehouseName=name, **kwargs)


@integrations.command(help=f'Setup a Snowflake integration. {WAREHOUSE_VERBIAGE}.')
@click.pass_obj
@click.option('--user', help='User with access to snowflake.', required=True)
@click.option('--account', help='Snowflake account name.', required=True)
@click.option('--warehouse', help='Name of the warehouse for the user.', required=True)
@click.option('--password', help=f'User\'s password. {PASSWORD_VERBIAGE}.', required=True,
              cls=AdvancedOptions, prompt_if_requested=True)
@click.option('--name', help='Friendly name for the warehouse.', required=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_snowflake(ctx, name, **kwargs):
    """
    Onboard a snowflake connection
    """
    WarehouseOnboardingService(config=ctx['config']).onboard_snowflake(warehouseName=name, **kwargs)


@integrations.command(help=f'Setup a BigQuery integration. {WAREHOUSE_VERBIAGE}.')
@click.pass_obj
@click.option('--key-file', help='JSON Key file.', type=click.Path(exists=True), required=True)
@click.option('--name', help='Friendly name for the warehouse.', required=True)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_bigquery(ctx, key_file, name, **kwargs):
    """
    Onboard a BigQuery connection
    """
    WarehouseOnboardingService(config=ctx['config']).onboard_bq(ServiceFile=key_file, warehouseName=name, **kwargs)


@integrations.command(help='Setup an integration that uses self-hosted credentials.')
@click.pass_obj
@click.option(
    '--connection-type',
    help='Type of connection.',
    required=True,
    type=click.Choice(GQL_TO_FRIENDLY_CONNECTION_MAP.keys(), case_sensitive=False),
    cls=AdvancedOptions,
    values_with_required_options=CONNECTION_TO_WAREHOUSE_TYPE_MAP.keys(),
    required_options_for_values=['name'],
)
@click.option(
    '--mechanism',
    help='Credential self-hosting mechanism.',
    required=True,
    type=click.Choice(SELF_HOSTING_MECHANISMS, case_sensitive=False),
    default=SECRETS_MANAGER_CREDENTIAL_MECHANISM,
)
@click.option('--key', help='Identifier for credentials within self-hosting mechanism.', required=True)
@click.option('--name', help='Friendly name for the warehouse. Required for warehouses.', required=False)
@add_common_options(ROLE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_self_hosted_credentials(ctx, mechanism, key, role, name, **kwargs):
    """
    Onboard a connection with self-hosted credentials
    """
    SelfHostedCredentialOnboardingService(config=ctx['config']).onboard_connection(
        self_hosting_mechanism=mechanism,
        self_hosting_key=key,
        assumable_role=role,
        warehouseName=name,
        **kwargs
    )


@integrations.command(help=f'Toggle S3 metadata events for Hive/Presto. {EVENT_VERBIAGE}.')
@click.pass_obj
@add_common_options(EVENT_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def toggle_hive_metadata_events(ctx, toggle):
    """
    Toggle s3 metadata events for a hive lake
    """
    EventsOnboardingService(config=ctx['config']).toggle_event_configuration(enable=toggle,
                                                                             connection_type=HIVE_MYSQL_CONNECTION_TYPE,
                                                                             event_type=S3_METADATA_EVENT_TYPE)


@integrations.command(help=f'Toggle S3 metadata events for Glue/Athena. {EVENT_VERBIAGE}.')
@click.pass_obj
@add_common_options(EVENT_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def toggle_glue_metadata_events(ctx, toggle):
    """
    Toggle s3 metadata events for a glue lake
    """
    EventsOnboardingService(config=ctx['config']).toggle_event_configuration(enable=toggle,
                                                                             connection_type=GLUE_CONNECTION_TYPE,
                                                                             event_type=S3_METADATA_EVENT_TYPE)


@integrations.command(help=f'Toggle S3 query log events. {EVENT_VERBIAGE}.')
@click.pass_obj
@add_common_options(EVENT_OPTIONS)
@add_common_options(ROLE_OPTIONS)
@click.option('--connection-type', help='Type of the integration.',
              type=click.Choice([HIVE_S3_CONNECTION_TYPE, PRESTO_S3_CONNECTION_TYPE], case_sensitive=True),
              required=False)
@click.option('--format-type', help='Query log format.',
              type=click.Choice(['hive-emr', 'custom'], case_sensitive=True),
              required=False)
@click.option('--source-format', help='Query log file format. Only required when "custom" is used.',
              type=click.Choice(['json', 'jsonl'], case_sensitive=True),
              required=False)
@click.option('--mapping-file',
              help='Mapping of expected to existing query log fields. Only required if "custom" is used.',
              type=click.Path(exists=True),
              required=False)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def toggle_ql_events(ctx, toggle, role, **kwargs):
    """
    Toggle s3 query log events for a lake
    """
    EventsOnboardingService(config=ctx['config']).toggle_event_configuration(enable=toggle,
                                                                             event_type=S3_QL_EVENT_TYPE,
                                                                             assumable_role=role,
                                                                             **kwargs)


@integrations.command(help=f'Setup a Tableau integration. {BI_VERBIAGE}.')
@click.pass_obj
@click.option('--token-name', help='Name for the personal access token.', cls=AdvancedOptions,
              mutually_exclusive_options=['password', 'username'], required_with_options=['token_value'],
              at_least_one_set=['password', 'username', 'token_name', 'token_value'])
@click.option('--token-value', help=f'Value for the personal access token. {PASSWORD_VERBIAGE}.', cls=AdvancedOptions,
              mutually_exclusive_options=['password', 'username'], required_with_options=['token_name'],
              prompt_if_requested=True)
@click.option('--password', help=f'Password for the service account. {PASSWORD_VERBIAGE}.', cls=AdvancedOptions,
              mutually_exclusive_options=['token_name', 'token_value'], required_with_options=['user'],
              prompt_if_requested=True)
@click.option('--user', help='Username for the service account.', cls=AdvancedOptions,
              mutually_exclusive_options=['token_name', 'token_value'], required_with_options=['password'])
@click.option('--site-name', help='The Tableau site name.', required=True)
@click.option('--server-name', help='The Tableau server name.', required=True)
@add_common_options(BI_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_tableau(ctx, user, **kwargs):
    """
    Onboard a tableau connection
    """
    ReportsOnboardingService(config=ctx['config']).onboard_tableau(username=user, **kwargs)


@integrations.command(help=f'Setup a Looker metadata integration. {BI_VERBIAGE}.')
@click.pass_obj
@click.option('--host-url', 'base_url', help='Looker host url.', required=True)
@click.option('--client-id', help='Looker client id.', required=True)
@click.option('--client-secret', help=f'Looker client secret (API key). {PASSWORD_VERBIAGE}.',
              required=True, cls=AdvancedOptions, prompt_if_requested=True)
@add_common_options(BI_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_looker(ctx, **kwargs):
    """
    Onboard a looker metadata connection
    """
    ReportsOnboardingService(config=ctx['config']).onboard_looker_metadata(**kwargs)


@integrations.command(help=f'Setup a Looker ML (git) integration. {BI_VERBIAGE}.')
@click.pass_obj
@click.option('--ssh-key', help='The ssh key for git ssh integrations.', required=False,
              type=click.Path(dir_okay=False, exists=True),
              cls=AdvancedOptions, mutually_exclusive_options=['token', 'username'])
@click.option('--repo-url', help='Repository URL as ssh://[user@]server/project.git or the shorter '
                                 'form [user@]server:project.git for ssh. For https, use https://server/project.git.',
              required=True)
@click.option('--token', help='Git Access Token to be used for Https instead of ssh key.', required=False,
              cls=AdvancedOptions, mutually_exclusive_options=['ssh-key'])
@click.option('--username', help='Git username to be used in conjunction with the access token. This is only '
                                 'required for BitBucket integrations.',
              cls=AdvancedOptions, required=False, mutually_exclusive_options=['ssh-key'])
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_looker_git(ctx, **kwargs):
    """
    Onboard a looker metadata connection
    """
    ReportsOnboardingService(config=ctx['config']).onboard_looker_git(**kwargs)


@integrations.command(help='List all active connections.', name='list')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def display_integrations(ctx):
    """
    Display active integrations
    """
    OnboardingStatusService(config=ctx['config']).display_integrations()


@integrations.command(help='Create an IAM role from a policy FILE. '
                           'The returned role ARN and external id should be used for adding lake assets.')
@click.pass_obj
@click.argument('file', type=click.Path(dir_okay=False, exists=True))
@click.option('--aws-profile', required=False,
              help='Override the AWS profile used by the CLI, which determines where the role is created. '
                   'This can be helpful when the account that manages the asset is not the same as the collector.')
def create_role(ctx, file, aws_profile):
    """
    Create a collector compatible role from the provided policy
    """
    CloudResourceService(config=ctx['config'], aws_profile_override=aws_profile).create_role(path_to_policy_doc=file)


@integrations.command(help='Update credentials for a connection. Only replaces/inserts the keys in changes by default.', name='update')
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click.option('--changes', help="""
              Credential key,value pairs as JSON.
              \b
              \n
              E.g. --changes '{"user":"Apollo"}'
              """, required=True, callback=validate_json_callback)
@click.option('--skip-validation', help='Skip validating credentials.', default=False, show_default=True, is_flag=True)
@click.option('--replace-all', help='Replace all credentials rather than just inserting/updating the keys in changes.', default=False, show_default=True, is_flag=True)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def update_credentials(ctx, connection_id, changes, skip_validation, replace_all):
    """
    Update credentials for a connection
    """
    ConnectionOperationsService(config=ctx['config']).update_credentials(
        connection_id=connection_id,
        changes=changes,
        should_validate=not skip_validation,
        should_replace=replace_all
    )


@integrations.command(help='Remove an existing connection. Deletes any associated jobs, monitors, etc.', name='remove')
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click.option('--no-prompt', help='Don\'t ask for confirmation.', default=False, show_default=True, is_flag=True)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def remove_connection(ctx, **kwargs):
    """
    Remove connection by ID
    """
    ConnectionOperationsService(config=ctx['config']).remove_connection(**kwargs)


@integrations.command(help='Retest an existing connection.', name='test')
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def echo_test_existing(ctx, **kwargs):
    """
    Tests an existing connection and echos results in pretty JSON.
    """
    ConnectionOperationsService(config=ctx['config']).echo_test_existing(**kwargs)


@integrations.command(help='Setup S3 event notifications for a lake.')
@click.pass_obj
@click.option('--bucket', 'bucket_name', required=True, help='Name of bucket to enable events for.')
@click.option('--prefix', default=None, callback=convert_empty_str_callback, required=False,
              help='Limit the notifications to objects starting with a prefix (e.g. \'data/\').')
@click.option('--suffix', default=None, callback=convert_empty_str_callback, required=False,
              help='Limit notifications to objects ending with a suffix (e.g. \'.csv\').')
@click.option('--topic-arn', default=None, callback=convert_empty_str_callback, required=False,
              help='Use an existing SNS topic (same region as the bucket). Creates a topic if one is not specified or '
                   'if an MCD topic does not already exist in the region.')
@click.option('--collector-aws-profile', required=False,
              help=f'Override the AWS profile use by the CLI for operations on SQS/Collector. {RESOURCE_VERBIAGE}.')
@click.option('--resource-aws-profile', 'bucket_aws_profile', required=False,
              help=f'Override the AWS profile use by the CLI for operations on S3/SNS. {RESOURCE_VERBIAGE}.')
@click.option('--event-type', required=True, default=CloudResourceService.MCD_EVENT_TYPE_MD, show_default=True,
              help='Type of event to setup.',
              type=click.Choice(CloudResourceService.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP.keys(), case_sensitive=True))
@click.option('--auto-yes', is_flag=True, help='Skip any interactive approval.', default=False, show_default=True)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_events(ctx, **kwargs):
    CloudResourceService(config=ctx['config'], aws_wrapper=None).add_events(**kwargs)


@integrations.command(help='Create an integration key. The resulting key id and secret will be printed to the console.')
@click.pass_obj
@click.option('--description',
              required=True,
              help='Key description.')
@click.option('--scope',
              required=True,
              type=click.Choice(IntegrationKeyScope.values(), case_sensitive=False),
              help='Key scope (integration the key can be used for).')
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def create_key(ctx, **kwargs):
    IntegrationKeyService(config=ctx['config']).create(**kwargs)


@integrations.command(help='Delete an integration key.')
@click.pass_obj
@click.option('--key-id', required=True, help='Integration key id.')
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete_key(ctx, **kwargs):
    IntegrationKeyService(config=ctx['config']).delete(**kwargs)


@integrations.command(help='List all integration keys.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def list_keys(ctx):
    IntegrationKeyService(config=ctx['config']).get_all()



import copy
import json
import secrets
import time
from collections import OrderedDict
from typing import Optional, Dict, Iterable, List, Union

import click
from box import Box, BoxList
from tabulate import tabulate

from montecarlodata.common.common import struct_match
from montecarlodata.common.data import EventProperties, AWSArn, DcResourceProperties
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors, prompt_connection
from montecarlodata.utils import AwsClientWrapper


class CloudResourceService:
    MCD_EVENT_TYPE_MD = 'metadata'
    MCD_EVENT_TYPE_QL = 'query-logs'
    MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP = {
        MCD_EVENT_TYPE_MD: 'MetadataEventQueue',
        MCD_EVENT_TYPE_QL: 'QueryLogEventQueue'
    }
    MCD_EVENT_Q_SENDER_ID = '__sender'

    _MCD_TOPIC_NAME_PATTERN = 'monte-carlo-data-{}-{}-events-topic'
    _MCD_TOPIC_POLICY_SID_PATTERN = 'monte-carlo-data-{}-{}-ID'
    _MCD_EVENT_ID_PATTERN = '{}{}{}-notification'

    _MCD_ROLE_PREFIX = 'monte-carlo-integration-role'
    _MCD_POLICY_PREFIX = 'monte-carlo-integration-cli-policy'
    _MCD_TAGS = [{'Key': 'MonteCarloData', 'Value': ''}]
    _LIST_EMR_FRIENDLY_HEADERS = ['Name', 'Id', 'State', 'LogUri']

    def __init__(self, config: Config, user_service: Optional[UserService] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None, aws_profile_override: Optional[str] = None,
                 aws_region_override: Optional[str] = None):
        self._abort_on_error = True
        self._profile = aws_profile_override or config.aws_profile
        self._region = aws_region_override or config.aws_region

        self._user_service = user_service or UserService(config=config)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=self._profile, region_name=self._region)

    @manage_errors
    def create_role(self, path_to_policy_doc: str) -> None:
        """
        Creates a DC compatible role from the provided policy doc
        """
        current_time = str(time.time())
        external_id = self._generate_random_token()
        role_name = f'{self._MCD_ROLE_PREFIX}-{current_time}'
        policy_name = f'{self._MCD_POLICY_PREFIX}-{current_time}'

        try:
            policy = json.dumps(self._read_json(path_to_policy_doc))

            # use the AWS account id of collector, which may not necessarily be account id running the CLI
            account_id = self._user_service.active_collector['customerAwsAccountId']
        except json.decoder.JSONDecodeError as err:
            complain_and_abort(f'The provided policy is not valid JSON - {err}')
        except KeyError as err:
            complain_and_abort(f'Missing expected property ({err}). The collector may not have been deployed before')
        else:
            trust_policy = self._generate_trust_policy(account_id=account_id, external_id=external_id)
            role_arn = self._aws_wrapper.create_role(role_name=role_name, trust_policy=trust_policy, tags=self._MCD_TAGS)
            click.echo(f"Created role with ARN - '{role_arn}' and external id - '{external_id}'.")

            self._aws_wrapper.attach_inline_policy(role_name=role_name, policy_name=policy_name, policy_doc=policy)
            click.echo(f'Success! Attached provided policy.')

    @staticmethod
    def _generate_trust_policy(account_id: str, external_id: str) -> str:
        """
        Generates a DC compatible trust policy
        """
        return json.dumps(
            {
                'Version': '2012-10-17',
                'Statement': [
                    {
                        'Effect': 'Allow',
                        'Principal': {
                            'AWS': f'arn:aws:iam::{account_id}:root'
                        },
                        'Action': 'sts:AssumeRole',
                        'Condition': {
                            'StringEquals': {
                                'sts:ExternalId': external_id
                            }
                        }
                    }
                ]
            }
        )

    @staticmethod
    def _generate_random_token(length: Optional[int] = 16) -> str:
        """
        Generates a random token (e.g. for the external id)
        """
        return secrets.token_urlsafe(length)

    @staticmethod
    def _read_json(path: str) -> Dict:
        """
        Reads a JSON file from the path.
        """
        with open(path) as file:
            return json.load(file)  # loads for the purpose of validating

    @manage_errors
    def list_emr_clusters(self, only_log_locations: Optional[bool] = False, created_after: Optional[str] = None,
                          states: Optional[List] = None, no_grid: Optional[bool] = False,
                          headers: Optional[str] = 'firstrow', table_format: Optional[str] = 'fancy_grid'):
        """
        Displays information about EMR cluster (name, id, state and log location). If only_log_locations is True,
        displays a deduplicated list of EMR log locations.
        """
        clusters = self._aws_wrapper.get_emr_cluster_details(created_after=created_after, states=states)
        if only_log_locations:
            self._list_emr_clusters_only_log_locations(clusters)
        elif no_grid:
            self._list_emr_clusters_details_no_wait(clusters)
        else:
            self._list_emr_cluster_details(clusters, headers=headers, table_format=table_format)

    def _list_emr_cluster_details(self, clusters: Iterable[Dict], headers: Optional[str] = 'firstrow',
                                  table_format: Optional[str] = 'fancy_grid'):
        rows = [
            self._emr_details_row(cluster) for cluster in clusters if cluster
        ]
        click.echo(tabulate([self._LIST_EMR_FRIENDLY_HEADERS] + rows, headers=headers, tablefmt=table_format))

    def _list_emr_clusters_details_no_wait(self, clusters: Iterable[Dict]):
        """
        Display information row by row, used when it can take long to wait for all data to format the table.
        """
        found = False
        for cluster in clusters:
            if not cluster:  # Should never happen.
                continue
            found = True
            # Prints line by line to show progress as the list may be long.
            click.echo(tabulate([self._emr_details_row(cluster)], tablefmt='plain'))
        if not found:
            click.echo(f'No clusters found')

    @staticmethod
    def _list_emr_clusters_only_log_locations(clusters: Iterable[Dict]):
        log_locations = set()
        for cluster in clusters:
            if not cluster:  # Should never happen.
                continue
            log_locations.add(cluster.get('LogUri'))
        if not log_locations:
            click.echo(f'No clusters found')
        else:
            for location in log_locations:
                click.echo(location)

    @staticmethod
    def _emr_details_row(cluster: Dict) -> List[str]:
        return [cluster['Name'], cluster['Id'], cluster['Status']['State'], cluster.get('LogUri')]

    @manage_errors
    def add_events(self, dc_id: Optional[str] = None, auto_yes: Optional[bool] = False, **kwargs) -> None:
        """
        Adds events for bucket in a repeatable manner (e.g. can be re-run on error in between steps) -
            1. Derives collection and resource related properties
            2. Creates topic or updates topic policy if necessary
            3. Updates (or sets) SQS policy with topic if not already set
            4. Subscribes queue to topic if not already done
            5. Creates notification if not already set

        All these steps are idempotent, so if if a permission or resource already exists it will not be duplicated.

        This is done via SDK rather than CloudFormation (CF) as CF does not support updating notification settings,
        SQS policies or topic policies for existing resources without using a custom resources (i.e. lambda functions),
        which is a less flexible variant of the actions performed here as it is bound by CF inputs and any steps and
        would still use an AWS SDK, without providing any of the common benefits of IaC. Additionally, cross-region
        and cross-account operations are non-trivial in CF, especially in a generic way without prior knowledge of the
        environment.

        Note - the only newly created resource (i.e. SNS topic) can be created externally and passed as an option.
        """
        click.echo('[Gathering collector and resource details]')
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._initialize_events(collector_props=collector, **kwargs)
        click.echo('[Creating or updating topic properties]')
        self._create_or_update_event_topic(event_details=event_details, skip_prompts=auto_yes)
        click.echo('[Updating event queue properties]')
        self._update_event_q(event_details=event_details, skip_prompts=auto_yes)
        click.echo('[Creating topic subscription]')
        self._create_topic_subscription(event_details=event_details, skip_prompts=auto_yes)
        click.echo('[Creating event notification]')
        self._create_event_notification(event_details=event_details, skip_prompts=auto_yes)
        click.echo('\nProcess complete. Have a nice day!')

    def _initialize_events(self, bucket_name: str, event_type: str, collector_props: Box,
                           collector_aws_profile: Optional[str] = None,
                           bucket_aws_profile: Optional[str] = None, **kwargs) -> EventProperties:
        """
        Collect (or lookup) any relevant event related properties and initialize any clients.

        Note - modifies collector_props by adding `outputs`.
        """
        collector_arn = AWSArn(id=collector_props.stackArn)
        collector_profile = collector_aws_profile or self._profile
        cust_resources_profile = bucket_aws_profile or self._profile

        # Create a temp client to lookup the region of the bucket (reduces number of prompts).
        temp_client = AwsClientWrapper(profile_name=cust_resources_profile, region_name=collector_arn.region)
        cust_resources_region = temp_client.get_bucket_location(bucket_name=bucket_name)

        # Create a client for collector resources and any customer resources.
        # This is necessary as the collector resources might be in a different account/region than customer resources.
        collection_client = AwsClientWrapper(profile_name=collector_profile, region_name=collector_arn.region)
        cust_resources_client = AwsClientWrapper(profile_name=cust_resources_profile, region_name=cust_resources_region)

        collector_props.outputs = collection_client.get_stack_outputs(stack_id=collector_props.stackArn)
        event_queue_arn = self._get_relevant_q(event_type, collector_props.outputs)

        return EventProperties(
            bucket_name=bucket_name,
            event_type=event_type,
            event_queue_arn=event_queue_arn,
            collection_region=collector_arn.region,
            cust_resources_region=cust_resources_region,
            collection_account_id=collector_arn.account,
            cust_account_id=cust_resources_client.get_caller_identity(),
            collection_client=collection_client,
            cust_resources_client=cust_resources_client,
            **kwargs
        )

    def _create_or_update_event_topic(self, event_details: EventProperties, skip_prompts: Optional[bool] = False) -> None:
        """
        Creates topic (if necessary) and/or updates policy to allow access from bucket (put) & event queue (subscribe)
        """
        new_statements = []
        b_topic_policy_id = self._MCD_TOPIC_POLICY_SID_PATTERN.format(event_details.event_type, event_details.bucket_name)
        q_topic_policy_id = self._MCD_TOPIC_POLICY_SID_PATTERN.format(event_details.event_type, event_details.event_queue_name)

        if event_details.topic_arn is None:
            # Create an event topic if an MC one does not already exist in the region for the queue type
            name = self._MCD_TOPIC_NAME_PATTERN.format(event_details.cust_resources_region, event_details.event_type)
            prompt = f'Create topic \'{name}\' in \'{event_details.cust_resources_region}\' ' \
                     f'({event_details.cust_account_id})? This is an idempotent operation.'
            prompt_connection(message=prompt, skip_prompt=skip_prompts)
            event_details.topic_arn = event_details.cust_resources_client.create_sns_topic(name=name)
        event_details.topic_name = self._validate_topic_arn(event_details=event_details).resource

        # Update policy if necessary
        topic_attributes = event_details.cust_resources_client.get_topic_attributes(arn=event_details.topic_arn)
        if not self._is_preexisting_topic_policy(topic_policy_id=b_topic_policy_id, topic_attributes=topic_attributes):
            new_statements.append(self._generate_sns_bucket_policy(sid=b_topic_policy_id, event_details=event_details))
        if not self._is_preexisting_topic_policy(topic_policy_id=q_topic_policy_id, topic_attributes=topic_attributes):
            new_statements.append(self._generate_sns_q_policy(sid=q_topic_policy_id, event_details=event_details))

        new_statement_count = len(new_statements)
        if new_statement_count > 0:
            click.echo(f'Adding statements (x{new_statement_count}) to \'{event_details.topic_name}\' -')
            click.echo(json.dumps(new_statements, indent=4))
            prompt_connection(message='Please confirm', skip_prompt=skip_prompts)
            event_details.cust_resources_client.set_topic_attributes(
                arn=event_details.topic_arn,
                name='Policy',
                value=self._extend_topic_policy(topic_attributes=topic_attributes, new_statements=new_statements)
            )
        else:
            click.echo(f'No policy changes required for \'{event_details.topic_name}\'')

    def _update_event_q(self, event_details: EventProperties, skip_prompts: Optional[bool] = False) -> None:
        """
        Create or update queue policy with topic
        """
        attributes = event_details.collection_client.get_q_attributes(
            name=event_details.event_queue_name,
            attributes=['Policy']
        ) or {}

        if not attributes.get('Attributes', {}).get('Policy'):
            # No policy set (e.g. is a new q). Create
            new_policy = self._generate_q_policy(event_details=event_details)
            click.echo(f'No existing policy found. Adding -\n{json.dumps(new_policy, indent=4)}')
        else:
            # Update existing if necessary
            existing_policy = Box(json.loads(attributes.Attributes.Policy))
            new_policy = self._extend_q_policy(existing_policy=existing_policy, event_details=event_details)
            if struct_match(existing_policy, new_policy):
                click.echo(f'Policy for \'{event_details.event_queue_name}\' requires no changes')
                return
            click.echo(f'Updating existing policy into -\n{json.dumps(new_policy, indent=4)}')

        prompt_connection(message='Please confirm', skip_prompt=skip_prompts)
        event_details.collection_client.set_q_attributes(
            name=event_details.event_queue_name,
            attributes=dict(Policy=json.dumps(new_policy))
        )

    def _create_event_notification(self, event_details: EventProperties, skip_prompts: Optional[bool] = False) -> None:
        """
        Creates event notification if one does not already exist for the topic
        """
        setting_id = self._MCD_EVENT_ID_PATTERN.format(
            AWSArn(id=event_details.topic_arn).resource.lower(),
            f'-{event_details.prefix}' if event_details.prefix else '',
            f'-{event_details.suffix}' if event_details.suffix else ''
        )

        # Get existing notification settings for bucket as AWS does not support inserting (only a put operation)
        existing_config = event_details.cust_resources_client.get_bucket_event_config(name=event_details.bucket_name)
        existing_config.pop('ResponseMetadata', None)
        message = 'Updating existing notifications into -' if existing_config else 'Creating notification -'

        new_config = self._merge_event_notifications(
            setting_id=setting_id,
            existing_config=existing_config,
            event_details=event_details
        )

        if struct_match(existing_config, new_config):
            click.echo(f'Notifications for \'{event_details.bucket_name}\' require no changes')
            return

        click.echo(f'{message}\n{json.dumps(new_config, indent=4)}')
        prompt_connection(message='Please confirm', skip_prompt=skip_prompts)
        event_details.cust_resources_client.set_bucket_event_config(
            name=event_details.bucket_name,
            notification_config=new_config
        )

    def _create_topic_subscription(self, event_details: EventProperties, skip_prompts: Optional[bool] = False) -> None:
        """
        Create topic subscription for endpoint if one for the endpoint already does not exist
        """
        subscriptions = list(event_details.cust_resources_client.list_topic_subscriptions(arn=event_details.topic_arn) or [])
        if subscriptions:
            for subscription in [sub for sublist in subscriptions for sub in sublist]:
                if subscription.get('Endpoint') == event_details.event_queue_arn:
                    click.echo(f'Subscription for \'{event_details.event_queue_name}\' already exists')
                    return
        click.echo(f'Creating subscription for \'{event_details.event_queue_name}\' on \'{event_details.topic_name}\'')
        prompt_connection(message='Please confirm', skip_prompt=skip_prompts)
        event_details.cust_resources_client.subscribe_to_topic(
            arn=event_details.topic_arn,
            endpoint=event_details.event_queue_arn,
            attributes=dict(RawMessageDelivery='true')
        )

    def _get_relevant_q(self, event_type: str, outputs: Union[Box, BoxList]) -> str:
        """
        Get queue arn from stack outputs based on the friendly key.

        Raise error if not found (e.g. events are not enabled)
        """
        if not event_type or event_type not in self.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP.keys():
            complain_and_abort(f'Event type was not defined or is invalid \'{event_type}\'.')

        for output in outputs:
            if output.OutputKey == self.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP[event_type]:
                return output.OutputValue
        complain_and_abort('Failed to find the relevant event queue. Are events enabled for this collector?')

    @staticmethod
    def _validate_topic_arn(event_details: EventProperties) -> AWSArn:
        """
        Sanity check ARN (e.g. if provided) is valid (format + region)
        """
        try:
            arn = AWSArn(id=event_details.topic_arn)
        except IndexError:
            complain_and_abort(f'SNS topic ARN \'{event_details.topic_arn}\' is not a valid.')
        else:
            if event_details.cust_resources_region != arn.region:
                complain_and_abort(f'Topic region ({arn.region}) does not match bucket '
                                   f'region ({event_details.cust_resources_region})')
            return arn

    @staticmethod
    def _is_preexisting_topic_policy(topic_policy_id: str, topic_attributes: Box) -> bool:
        """
        Check if a policy for this bucket already exists. Returns true if it does.
        """
        for statement in Box(json.loads(topic_attributes.Attributes.Policy)).Statement:
            if statement.get('Sid') and statement.Sid == topic_policy_id:
                return True
        return False

    @staticmethod
    def _extend_topic_policy(topic_attributes: Box, new_statements: List[Dict]) -> str:
        """
        Merge new statements with existing
        """
        policy = Box(json.loads(topic_attributes.Attributes.Policy))
        policy.Statement.extend(new_statements)
        return json.dumps(policy)

    @staticmethod
    def _extend_q_policy(existing_policy: Box, event_details: EventProperties) -> Dict:
        """
        Merge new statements with existing
        """
        new_policy = Box(copy.deepcopy(existing_policy))
        for statement in new_policy.Statement:
            if statement.get('Sid') and statement.Sid == CloudResourceService.MCD_EVENT_Q_SENDER_ID:
                source_arn = statement.get('Condition', {}).get('ArnLike', {}).get('aws:SourceArn')
                if source_arn:
                    # Source ARN can be a list or string :/
                    if isinstance(source_arn, List):
                        source_arn.append(event_details.topic_arn)
                    else:
                        source_arn = [source_arn, event_details.topic_arn]
                    # Remove duplicates, while still retaining order/type for computing diff.
                    statement.Condition.ArnLike['aws:SourceArn'] = list(OrderedDict.fromkeys(source_arn).keys())
                    if len(statement.Condition.ArnLike['aws:SourceArn']) == 1:
                        statement.Condition.ArnLike['aws:SourceArn'] = statement.Condition.ArnLike['aws:SourceArn'][0]
                    return new_policy
        # Existing policy was create outside of tooling / different from steps outlined in documentation
        complain_and_abort('Queue policy exists, but not in a supported format')

    def _merge_event_notifications(self, setting_id: str, existing_config: Dict, event_details: EventProperties) -> Dict:
        """
        Merge new config with existing based on if and how topic config was set
        """
        config = copy.deepcopy(existing_config)
        if config.get('TopicConfigurations'):
            # existing topic notification configurations
            for topic in config['TopicConfigurations']:
                if topic.get('Id') == setting_id:
                    return config  # No changes required; can short circuit
            config['TopicConfigurations'].append(
                self._generate_bucket_to_topic_config(
                    setting_id=setting_id,
                    event_details=event_details
                )
            )
        else:
            # existing queue or lambda notification configurations
            config['TopicConfigurations'] = [
                self._generate_bucket_to_topic_config(
                    setting_id=setting_id,
                    event_details=event_details
                )
            ]
        return config

    @staticmethod
    def _generate_bucket_to_topic_config(setting_id: str, event_details: EventProperties) -> Dict:
        return dict(
            Id=setting_id,
            TopicArn=event_details.topic_arn,
            Events=['s3:ObjectCreated:*', 's3:ObjectRemoved:*'],
            Filter=dict(
                Key=dict(
                    FilterRules=[
                        dict(Name='Prefix', Value=event_details.prefix or ''),
                        dict(Name='Suffix', Value=event_details.suffix or '')
                    ]
                )
            )
        )

    @staticmethod
    def _generate_sns_bucket_policy(sid: str, event_details: EventProperties) -> Dict:
        return {
            'Sid': sid,
            'Effect': 'Allow',
            'Principal': {
                'AWS': '*'
            },
            'Action': 'SNS:Publish',
            'Resource': event_details.topic_arn,
            'Condition': {
                'StringEquals': {
                    'aws:SourceArn': event_details.bucket_arn
                }
            }
        }

    @staticmethod
    def _generate_sns_q_policy(sid: str, event_details: EventProperties) -> Dict:
        return {
            'Sid': sid,
            'Effect': 'Allow',
            'Principal': {
                'AWS': event_details.collection_account_id
            },
            'Action': 'sns:Subscribe',
            'Resource': event_details.topic_arn
        }

    @staticmethod
    def _generate_q_policy(event_details: EventProperties):
        return {
            'Version': '2008-10-17',
            'Statement': list(filter(None, [
                {
                    'Sid': '__owner',
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': f'arn:aws:iam::{event_details.collection_account_id}:root'
                    },
                    'Action': 'SQS:*',
                    'Resource': event_details.event_queue_arn
                },
                {
                    'Sid': CloudResourceService.MCD_EVENT_Q_SENDER_ID,
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': '*'
                    },
                    'Action': 'SQS:SendMessage',
                    'Resource': event_details.event_queue_arn,
                    'Condition': {
                        'ArnLike': {
                            'aws:SourceArn': [
                                event_details.topic_arn
                            ]
                        }
                    }
                } if event_details.topic_arn else None
            ]))
        }

    def get_and_validate_active_collector(self, dc_id: Optional[str] = None) -> Box:
        """
        Get a specific collector and validate that it is active.
        """
        collector = self._user_service.get_collector(dc_id=dc_id)
        if not collector.stackArn or not collector.active:
            complain_and_abort('Cannot setup events for an inactive collector.')
        return collector

    def get_dc_resource_props(self, *, collector_props: Box,
                              collector_aws_profile: Optional[str] = None,
                              resource_aws_profile: Optional[str] = None,
                              resource_aws_region: Optional[str] = None, **_) -> DcResourceProperties:
        """
        Collect (or lookup) DC and resource properties. Note - modifies collector_props by adding `outputs` and `parameters`.

        Multiple clients are created as collector resources might be in a different account/region than customer resources.
        """
        collector_arn = AWSArn(id=collector_props.stackArn)
        collector_profile = collector_aws_profile or self._profile
        cust_resources_profile = resource_aws_profile or self._profile

        collection_region = collector_arn.region
        resources_region = resource_aws_region or collection_region

        collection_client = AwsClientWrapper(profile_name=collector_profile, region_name=collection_region)
        resources_client = AwsClientWrapper(profile_name=cust_resources_profile, region_name=resources_region)

        collector_props.outputs = self._parse_stack_prop_list(
            key='OutputKey',
            val='OutputValue',
            struct=collection_client.get_stack_outputs(stack_id=collector_props.stackArn)
        )

        collector_props.parameters = self._parse_stack_prop_list(
            key='ParameterKey',
            val='ParameterValue',
            struct=collection_client.get_stack_parameters(stack_id=collector_props.stackArn)
        )

        return DcResourceProperties(
            collector_arn=collector_arn,
            collector_props=collector_props,
            collection_region=collection_region,
            resources_region=resources_region,
            collection_client=collection_client,
            resources_client=resources_client
        )

    @staticmethod
    def _parse_stack_prop_list(key: str, val: str, struct: Union[List[Dict], Box]) -> Box:
        """
        Parse stack props into an easier to use collapsed format
        """
        return Box({row[key]: row[val] for row in struct}, camel_killer_box=True)


import json
from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import LetterCase
from dataclasses_json import dataclass_json
from marshmallow import fields

from montecarlodata.iac.utils import field_spec
from montecarlodata.settings import DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION, DEFAULT_INCLUDE_PATTERNS, DEFAULT_EXCLUDE_PATTERNS


@dataclass_json
@dataclass
class ProjectConfig:
    version: Optional[int] = field_spec(fields.Int(required=False), default_factory=lambda: DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION)
    default_resource: Optional[str] = field_spec(fields.Str(required=False))
    include_file_patterns: Optional[List[str]] = field_spec(
        fields.List(fields.Str(required=True)),
        default_factory=lambda: DEFAULT_INCLUDE_PATTERNS)
    exclude_file_patterns: Optional[List[str]] = field_spec(
        fields.List(fields.Str(required=True)),
        default_factory=list)

    def __post_init__(self):
        self.exclude_file_patterns = list(set(self.exclude_file_patterns + DEFAULT_EXCLUDE_PATTERNS))


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ResourceModification:
    type: str
    description: str
    resource_as_json: str

    def __post_init__(self):
        self.resource = json.loads(self.resource_as_json)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ConfigTemplateUpdateResponse:
    resource_modifications: List[ResourceModification]
    errors_as_json: Optional[str] = None
    changes_applied: bool = False

    def __post_init__(self):
        self.errors = json.loads(self.errors_as_json)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ConfigTemplateDeleteResponse:
    num_deleted: int
    changes_applied: bool = False


import base64
import ipaddress
import json
from collections import Mapping
from functools import wraps
from typing import Optional, Dict, List, Any

from box import Box
from pycarlo.core import Client, Session

from montecarlodata.config import Config


def normalize_gql(field: str) -> Optional[str]:
    if field:
        return field.replace('_', '-').lower()


def read_as_base64(path: str) -> bytes:
    with open(path, 'rb') as fp:
        return base64.b64encode(fp.read())


def read_as_json(path: str) -> Dict:
    with open(path) as file:
        return json.load(file)


def read_as_json_string(path: str) -> str:
    """"Read and validate JSON file"""
    return json.dumps(read_as_json(path))


def struct_match(s1: Dict, s2: Dict) -> bool:
    return json.dumps(s1, sort_keys=True) == json.dumps(s2, sort_keys=True)


def boxify(use_snakes: Optional[bool] = False, default_box_attr: Optional[Any] = object()):
    """
    Convenience decorator to convert a dict into Box for ease of use.

    Set `use_snakes` to convert camelCase to snake_case. Use `default_box_attr` to set a default value.
    """
    def _boxify(func):
        @wraps(func)
        def _impl(self, *args, **kwargs):
            dict_ = func(self, *args, **kwargs)
            if dict_ and isinstance(dict_, Mapping):
                return Box(dict_, camel_killer_box=use_snakes, default_box_attr=default_box_attr)
            return dict_
        return _impl
    return _boxify


def chunks(lst: List, n: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def is_overlap(r1: str, r2: str) -> bool:
    """Check if two CIDR ranges overlap"""
    r1 = ipaddress.ip_network(r1)
    r2 = ipaddress.ip_network(r2)
    return r1.overlaps(r2) or r2.overlaps(r1)


def create_mc_client(ctx: Dict) -> Client:
    config: Config = ctx['config']
    mc_client = Client(
        session=Session(
            mcd_id=config.mcd_id,
            mcd_token=config.mcd_token
        )
    )
    return mc_client


import os
from pathlib import Path

"""Environmental configuration"""

# Verbose error logging
MCD_VERBOSE_ERRORS = os.getenv('MCD_VERBOSE_ERRORS', False) in (True, 'true', 'True')

# MCD API endpoint
MCD_API_ENDPOINT = os.getenv('MCD_API_ENDPOINT', 'https://api.getmontecarlo.com/graphql')

# MCD ID header (for use in local development and testing)
MCD_USER_ID_HEADER = os.getenv('MCD_USER_ID_HEADER')

# MCD API ID. Overwrites config-file
MCD_DEFAULT_API_ID = os.getenv('MCD_DEFAULT_API_ID')

# MCD API Token. Overwrites config-file
MCD_DEFAULT_API_TOKEN = os.getenv('MCD_DEFAULT_API_TOKEN')

# dbt cloud API token
DBT_CLOUD_API_TOKEN = os.getenv('DBT_CLOUD_API_TOKEN')

# dbt cloud account ID
DBT_CLOUD_ACCOUNT_ID = os.getenv('DBT_CLOUD_ACCOUNT_ID')


"""Tool Defaults"""

# Default profile to be used
DEFAULT_PROFILE_NAME = 'default'

# Default path where any configuration files are written
DEFAULT_CONFIG_PATH = os.path.join(str(Path.home()), '.mcd')

# Default region where data collector is deployed
DEFAULT_AWS_REGION = 'us-east-1'

"""Internal Use"""

# File name for profile configuration
PROFILE_FILE_NAME = 'profiles.ini'

# Configuration sub-command
CONFIG_SUB_COMMAND = 'configure'

# Help flag of arguments and options
HELP_FLAG = '--help'

# Option file flag
OPTION_FILE_FLAG = '--option-file'

# Value (if entered) to show prompt with hidden input
SHOW_PROMPT_VALUE = '-1'

"""Monitors as code configs"""

DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION = 1

PROJECT_CONFIG_FILENAME = 'montecarlo.yml'

TARGET_DIRECTORY_NAME = 'target'

DEFAULT_INCLUDE_PATTERNS = [
    '**/*.yaml',
    '**/*.yml',
]

DEFAULT_EXCLUDE_PATTERNS = [
    'target/*',
    'dbt_project.yml',
    'montecarlo.yml'
]



import json
from typing import Optional, Dict

from montecarlodata.iac.schemas import ConfigTemplateUpdateResponse, ConfigTemplateDeleteResponse
from montecarlodata.queries.iac import CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE, DELETE_MONTE_CARLO_CONFIG_TEMPLATE
from montecarlodata.utils import GqlWrapper


class MonteCarloConfigTemplateClient:
    def __init__(self, gql_wrapper: GqlWrapper):
        self._gql_wrapper = gql_wrapper

    def apply_config_template(self,
                              namespace: str,
                              config_template_as_dict: Dict,
                              resource: Optional[str] = None,
                              dry_run: bool = False) -> ConfigTemplateUpdateResponse:
        response = self._gql_wrapper.make_request_v2(
            query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE,
            operation='createOrUpdateMonteCarloConfigTemplate',
            variables=dict(
                namespace=namespace,
                configTemplateJson=json.dumps(config_template_as_dict),
                dryRun=dry_run,
                resource=resource
            )
        )

        return ConfigTemplateUpdateResponse.from_dict(response.data['response'])

    def delete_config_template(self,
                               namespace: str,
                               dry_run: bool = False) -> ConfigTemplateDeleteResponse:
        response = self._gql_wrapper.make_request_v2(
            query=DELETE_MONTE_CARLO_CONFIG_TEMPLATE,
            operation='deleteMonteCarloConfigTemplate',
            variables=dict(
                namespace=namespace,
                dryRun=dry_run,
            )
        )

        return ConfigTemplateDeleteResponse.from_dict(response.data['response'])



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD, \
    EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD, EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD, HIVE_MYSQL_CONNECTION_TYPE, \
    HIVE_S3_CONNECTION_TYPE, HIVE_SQL_CONNECTION_TYPE, QL_JOB_TYPE, AWS_RDS_CA_CERT, PRESTO_CATALOG_KEY, \
    HIVE_GET_PARTS_KEY, HIVE_MAX_PARTS_KEY, HIVE_MAX_PARTS_DEFAULT_VALUE
from montecarlodata.queries.onboarding import TEST_DATABASE_CRED_MUTATION, TEST_S3_CRED_MUTATION, \
    TEST_HIVE_SQL_CRED_MUTATION


class HiveOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_hive_mysql(self, **kwargs) -> None:
        """
        Onboard a hive-mysql connection by validating and adding a connection
        """
        if kwargs['use_ssl']:
            kwargs['ssl_options'] = {'ca': AWS_RDS_CA_CERT}
            kwargs.pop('use_ssl')

        catalog = kwargs.pop('catalog', None)
        job_limits = {HIVE_GET_PARTS_KEY: True, HIVE_MAX_PARTS_KEY: HIVE_MAX_PARTS_DEFAULT_VALUE}

        if catalog:
            job_limits[PRESTO_CATALOG_KEY] = catalog

        kwargs['connectionType'] = HIVE_MYSQL_CONNECTION_TYPE
        self.onboard(validation_query=TEST_DATABASE_CRED_MUTATION,
                     validation_response=EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
                     connection_type=HIVE_MYSQL_CONNECTION_TYPE, job_limits=job_limits, **kwargs)

    @manage_errors
    def onboard_hive_s3(self, **kwargs) -> None:
        """
        Onboard a presto-s3 connection by validating and adding a connection
        """
        kwargs['connectionType'] = HIVE_S3_CONNECTION_TYPE
        self.onboard(validation_query=TEST_S3_CRED_MUTATION,
                     validation_response=EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD,
                     connection_type=HIVE_S3_CONNECTION_TYPE, job_types=QL_JOB_TYPE, **kwargs)

    @manage_errors
    def onboard_hive_sql(self, **kwargs) -> None:
        """
        Onboard a hive-sql connection by validating and adding a connection
        """
        self.onboard(validation_query=TEST_HIVE_SQL_CRED_MUTATION,
                     validation_response=EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD,
                     connection_type=HIVE_SQL_CONNECTION_TYPE, **kwargs)



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD, \
    EXPECTED_PRESTO_S3_GQL_RESPONSE_FIELD, PRESTO_SQL_CONNECTION_TYPE, PRESTO_S3_CONNECTION_TYPE, QL_JOB_TYPE, \
    PRESTO_CERT_PREFIX
from montecarlodata.queries.onboarding import TEST_PRESTO_CRED_MUTATION, TEST_S3_CRED_MUTATION


class PrestoOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_presto_sql(self, **kwargs) -> None:
        """
        Onboard a presto-sql connection by validating and adding a connection.
        Also, optionally uploads a certificate to the DC bucket.
        """
        self.handle_cert(cert_prefix=PRESTO_CERT_PREFIX, options=kwargs)
        self.onboard(validation_query=TEST_PRESTO_CRED_MUTATION,
                     validation_response=EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD,
                     connection_type=PRESTO_SQL_CONNECTION_TYPE, **kwargs)

    @manage_errors
    def onboard_presto_s3(self, **kwargs) -> None:
        """
        Onboard a presto-s3 connection by validating and adding a connection
        """
        kwargs['connectionType'] = PRESTO_S3_CONNECTION_TYPE
        self.onboard(validation_query=TEST_S3_CRED_MUTATION,
                     validation_response=EXPECTED_PRESTO_S3_GQL_RESPONSE_FIELD,
                     connection_type=PRESTO_S3_CONNECTION_TYPE, job_types=QL_JOB_TYPE, **kwargs)



"""
Common commands
"""
import click

from montecarlodata.tools import convert_uuid_callback

DISAMBIGUATE_DC_OPTIONS = [
    click.option('--collector-id', 'dc_id', required=False, type=click.UUID, callback=convert_uuid_callback,
                 help='ID for the data collector. To disambiguate accounts with multiple collectors.'),
]



from dataclasses import dataclass
from typing import Optional, Dict, List

from box import Box

from montecarlodata.common.common import boxify
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.queries.user import GET_USER_QUERY
from montecarlodata.utils import GqlWrapper


@dataclass
class User:
    first_name: Optional[str] = None
    last_name: Optional[str] = None


class UserService:
    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None):
        self._config = config
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)

        self._user = self._request_wrapper.make_request(GET_USER_QUERY)  # get user info

    @property
    def user(self) -> User:
        """
        Returns basic user properties
        """
        return User(first_name=self._user['getUser'].get('firstName'), last_name=self._user['getUser'].get('lastName'))

    @property
    def account(self) -> Dict:
        """
        Get account details
        """
        return self._user['getUser']['account']

    @property
    def collectors(self) -> Dict:
        """
        Get collectors in the account
        """
        return self.account.get('dataCollectors', [{}])

    @property
    def active_collector(self) -> Dict:
        """
        Get active collector from collectors.

        Errors out on accounts with > 1 collector. Legacy. Do not use. See `get_collector`
        """
        return self.collectors[self._get_active_collector()]

    @property
    def warehouses(self) -> List[Dict]:
        """
        Get warehouses attached to the account
        """
        return self.account.get('warehouses')

    @property
    def bi_containers(self) -> List[Dict]:
        """
        Get bi attached to the account
        """
        return self.account.get('bi')

    @property
    def tableau_accounts(self) -> Dict:
        """
        Get tableau connections attached to the account
        """
        return self.account.get('tableauAccounts')

    @property
    def active_collection_regions(self) -> List[str]:
        """
        Get a list of active collection regions
        """
        return self.account.get('activeCollectionRegions')

    @boxify()
    def get_collector(self, dc_id: Optional[str] = None) -> Box:
        """
        Get a specific collector

        This is only necessary for client only ops on a specific collector. APIs handle this disambiguation already.
        """
        num_of_collectors = len(self.collectors)
        if num_of_collectors == 0:
            complain_and_abort('No collector found.')

        if dc_id:
            for collector in self.collectors:
                if dc_id == collector['uuid']:
                    return collector
            complain_and_abort(f'Collector with ID \'{dc_id}\' not found.')
        if num_of_collectors > 1:
            complain_and_abort('Multiple data collectors found. Please specify a collector.')
        return self.collectors[0]

    def _get_active_collector(self) -> Optional[int]:
        """
        Get active collector - currently only one active collector per account is supported. Abort if None are found
        """
        if len(self.collectors) > 1:
            complain_and_abort('This option is only supported in accounts with one collector.')

        for idx, collector in enumerate(self.collectors):
            if collector.get('active'):
                return idx
        complain_and_abort('No active collector found')



import contextlib
from dataclasses import field
from typing import Dict

from box import Box
from dataclasses_json import config


def field_spec(mm_field, default_factory=None, encoder=None, decoder=None):
    """
    dataclasses-json allows configuring a Marshmallow field which will be used for schema validation.
    See https://github.com/lidatong/dataclasses-json#overriding--extending
    Syntax involves a lot of boilerplate, so trying to minimize by using this function.
    """
    f = field(metadata=config(
        mm_field=mm_field,
        encoder=encoder,
        decoder=decoder
    ))

    # The following makes all of our dataclass fields optional.
    # We are enforcing required fields using marshmallow validators, so let's simplify and make
    # all the dataclass fields optional.
    if default_factory:
        f.default_factory = default_factory
    else:
        f.default = None

    return f

def is_dbt_schema(yaml_as_dict: Box) -> bool:
    with contextlib.suppress(KeyError):
        return isinstance(yaml_as_dict.models, list)
    return False

def has_montecarlo_property(yaml_as_dict: Dict) -> bool:
    return 'montecarlo' in yaml_as_dict


from typing import Optional, List, Dict

import click
from tabulate import tabulate

from montecarlodata.common.common import normalize_gql
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.fields import GQL_TO_FRIENDLY_CONNECTION_MAP


class OnboardingStatusService:
    _INTEGRATION_FRIENDLY_HEADERS = ['Integration', 'ID', 'Created on (UTC)']

    def __init__(self, config: Config, user_service: Optional[UserService] = None):
        self._abort_on_error = True
        self._user_service = user_service or UserService(config=config)

    @manage_errors
    def display_integrations(self, headers: Optional[str] = 'firstrow', table_format: Optional[str] = 'fancy_grid'):
        """
        Display active integrations in an easy to read table. E.g.
        ╒══════════════════╤══════════════════════════════════════╤══════════════════════════════════╕
        │ Integration      │ ID                                   │ Created on (UTC)                 │
        ╞══════════════════╪══════════════════════════════════════╪══════════════════════════════════╡
        │ Hive (metastore) │ 1023721b-e639-44ed-9838-e1f669d3fc85 │ 2020-05-09T01:49:52.806602+00:00 │
        ╘══════════════════╧══════════════════════════════════════╧══════════════════════════════════╛
        """
        table = [self._INTEGRATION_FRIENDLY_HEADERS]
        for integration in (self._user_service.warehouses or [{}]) + (self._user_service.bi_containers or [{}]):
            table += self._build_table_record(connections=integration.get('connections'))
        click.echo(tabulate(table, headers=headers, tablefmt=table_format))

    @staticmethod
    def _build_table_record(connections: List[Dict]) -> List[List[str]]:
        table = []
        for connection in connections or []:
            table.append(
                [
                    GQL_TO_FRIENDLY_CONNECTION_MAP.get(normalize_gql(connection['type']), connection['type']),
                    connection['uuid'],
                    connection['createdOn']
                ]
            )  # order by the friendly headers, defaulting to gql response if not found
        return table



import click

import montecarlodata.settings as settings
from montecarlodata.insights.commands import insights
from montecarlodata.collector.commands import collectors
from montecarlodata.dataimport.commands import import_subcommand
from montecarlodata.common.user import UserService
from montecarlodata.config import ConfigManager
from montecarlodata.discovery.commands import discovery
from montecarlodata.iac.commands import monitors
from montecarlodata.integrations.commands import integrations


@click.group(help='Monte Carlo\'s CLI.')
@click.option('--profile', default=settings.DEFAULT_PROFILE_NAME,
              help='Specify an MCD profile name. Uses default otherwise.')
@click.option('--config-path', default=settings.DEFAULT_CONFIG_PATH, type=click.Path(dir_okay=True),
              help=f'Specify path where to look for config file. Uses {settings.DEFAULT_CONFIG_PATH} otherwise.')
@click.version_option()
@click.pass_context
def entry_point(ctx, profile, config_path):
    """
    Entry point for all subcommands and options. Reads configuration and sets as context, except when configuring
    or getting help.
    """
    if ctx.invoked_subcommand != settings.CONFIG_SUB_COMMAND and settings.HELP_FLAG not in click.get_os_args():
        config = ConfigManager(profile_name=profile, base_path=config_path).read()
        if not config:
            ctx.abort()
        ctx.obj = {'config': config}


@click.command(help='Configure the CLI.')
@click.option('--profile-name', required=False, help='Specify a profile name for configuration.',
              default=settings.DEFAULT_PROFILE_NAME)
@click.option('--config-path', required=False, help='Specify path where to look for config file.',
              default=settings.DEFAULT_CONFIG_PATH, type=click.Path(dir_okay=True))
@click.option('--mcd-id', prompt='Key ID', help='Monte Carlo token user ID.')
@click.option('--mcd-token', prompt='Secret', help='Monte Carlo token value.', hide_input=True)
@click.option('--aws-profile', prompt='AWS profile name',
              help='AWS named profile with access to the DC stack and resources (e.g. CloudFormation and S3).',
              default='')
@click.option('--aws-region', prompt='AWS region', help='AWS region where DC is deployed.',
              default=settings.DEFAULT_AWS_REGION)
def configure(profile_name, config_path, mcd_id, mcd_token, aws_profile, aws_region):
    """
    Special subcommand for configuring the CLI
    """
    ConfigManager(profile_name=profile_name, base_path=config_path).write(mcd_id=mcd_id,
                                                                          mcd_token=mcd_token,
                                                                          aws_profile=aws_profile,
                                                                          aws_region=aws_region)


@click.command(help='Validate that the CLI can Connect to Monte Carlo.')
@click.pass_obj
def validate(ctx):
    """
    Special subcommand for validating the CLI was correctly configured
    """
    click.echo(f"Hi, {UserService(config=ctx['config']).user.first_name}! All is well.")


entry_point.add_command(integrations)
entry_point.add_command(configure)
entry_point.add_command(validate)
entry_point.add_command(collectors)
entry_point.add_command(discovery)
entry_point.add_command(monitors)
entry_point.add_command(import_subcommand)
entry_point.add_command(insights)


# to allow this to be run as a script within an IDE (for debugging)
if __name__ == '__main__':
    entry_point()


import click
from enum import Enum
from montecarlodata.common.data import ConnectionType
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.queries.keys import Queries
from montecarlodata.utils import GqlWrapper
from tabulate import tabulate
from typing import Dict, List, Optional


class IntegrationKeyScope(Enum):
    Spark = 'spark'

    @classmethod
    def values(cls) -> List[str]:
        return list(map(lambda s: s.name, cls))


class IntegrationKeyService:

    _table_headers = [
        'Id',
        'Description',
        'Scope',
        'Created',
        'Created By'
    ]

    def __init__(self,
                 config: Config,
                 gql: Optional[GqlWrapper] = None,
                 user_service: Optional[UserService] = None):

        self._gql = gql or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self._user_service = user_service or UserService(config=config)

        # used by @manage_errors decorator
        self._abort_on_error = True

    @manage_errors
    def create(self, description: str, scope: str):
        response = self._gql.make_request_v2(
            query=Queries.create.query,
            variables=self._resolve_variables(description, scope),
            operation=Queries.create.operation
        )

        click.echo(f'Key id: {response.data.key.id}')
        click.echo(f'Key secret: {response.data.key.secret}')

    def _resolve_variables(self, description: str, scope: str) -> Dict:
        variables = {
            'description': description,
            'scope': scope
        }

        if scope.lower() == IntegrationKeyScope.Spark.value:
            variables['warehouseIds'] = [self._resolve_lake_warehouse_id()]

        return variables

    def _resolve_lake_warehouse_id(self) -> str:
        lake_warehouse_ids = [w['uuid']
                              for w in self._user_service.warehouses
                              if w['connectionType'] == ConnectionType.DataLake.value]

        num_lakes = len(lake_warehouse_ids)
        if num_lakes == 0:
            complain_and_abort('Unable to resolve data lake connection: no lake connection found.')
        elif num_lakes > 1:
            complain_and_abort('Unable to resolve data lake connection: multiple lake connections found.')
        else:
            return lake_warehouse_ids[0]

    @manage_errors
    def delete(self, key_id: str):
        response = self._gql.make_request_v2(
            query=Queries.delete.query,
            variables = {
                'keyId': key_id
            },
            operation=Queries.delete.operation
        )

        if response.data.deleted:
            click.echo('Key has been deleted.')
        else:
            click.echo('Key was not deleted.')

    @manage_errors
    def get_all(self, table_format: Optional[str] = 'fancy_grid'):
        response = self._gql.make_request_v2(
            query=Queries.get_all.query,
            operation=Queries.get_all.operation
        )

        data = [[
            key.id,
            key.description,
            key.scope,
            key.createdTime,
            f'{key.createdBy.firstName} {key.createdBy.lastName}'
        ] for key in response.data]

        table = tabulate(data, headers=self._table_headers, tablefmt=table_format)
        click.echo(table)



from typing import Optional

import click

from montecarlodata.common.common import read_as_json_string
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD
from montecarlodata.queries.onboarding import TOGGLE_EVENT_MUTATION


class EventsOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def toggle_event_configuration(self, **kwargs) -> Optional[bool]:
        """
        Toggle event configuration (effectively onboarding it if enable is set to true)
        """
        num_of_warehouses = len(self._user_service.warehouses)

        if num_of_warehouses == 1:
            kwargs['dwId'] = self._user_service.warehouses[0]['uuid']
            mapping_file = kwargs.pop('mapping_file', None)
            if mapping_file:
                kwargs['mapping'] = read_as_json_string(mapping_file)
            variables = self._request_wrapper.convert_snakes_to_camels(kwargs)
            status = self._request_wrapper.make_request(query=TOGGLE_EVENT_MUTATION, variables=variables).get(
                EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD, {}).get('success')
            if status:
                click.echo(f"Success! {'Enabled' if kwargs['enable'] else 'Disabled'} events.")
                return True
            complain_and_abort('Failed to toggle events!')
        complain_and_abort(f'Events only supports one warehouse. Found {num_of_warehouses}')



import json
import os
from typing import Optional, Dict, List

import click
from tabulate import tabulate

from montecarlodata.common.data import ConnectionOptions, ValidationResult, OnboardingConfiguration
from montecarlodata.common.common import normalize_gql
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, prompt_connection
from montecarlodata.integrations.onboarding.fields import S3_CERT_MECHANISM, GQL_TO_FRIENDLY_CONNECTION_MAP, \
    DATA_LAKE_WAREHOUSE_TYPE, EXPECTED_ADD_CONNECTION_RESPONSE_FIELD, CONFIRM_CONNECTION_VERBIAGE, VALIDATIONS_FAILED_VERBIAGE, \
    SKIP_ADD_CONNECTION_VERBIAGE, CONNECTION_TEST_SUCCESS_VERBIAGE, CONNECTION_TEST_FAILED_VERBIAGE, \
    ADD_CONNECTION_FAILED_VERBIAGE, ADD_CONNECTION_SUCCESS_VERBIAGE
from montecarlodata.queries.onboarding import ADD_CONNECTION_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class BaseOnboardingService:
    _VALIDATION_STATUS_HEADERS = ['Validation', 'Result', 'Details']
    _VALIDATION_PASSED = 'Passed'
    _VALIDATION_FAILED = 'Failed'

    def __init__(self, config: Config, user_service: Optional[UserService] = None,
                 request_wrapper: Optional[GqlWrapper] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None):

        self._abort_on_error = True  # Aborts methods with deco on unhandled error
        self._dc_outputs = None

        self._user_service = user_service or UserService(config=config)
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=config.aws_profile,
                                                            region_name=config.aws_region)

    def onboard(self, **kwargs):
        """
        Convenience wrapper to validate and add a connection.
        """
        onboarding_config = OnboardingConfiguration(**kwargs)  # not passed as arg to avoid having to change all clients
        onboarding_config.connection_options = self._build_connection_options(**onboarding_config.connection_options)

        onboarding_config.warehouse_name = onboarding_config.connection_options.monolith_base_payload.pop('warehouseName', None)
        onboarding_config.warehouse_type = onboarding_config.connection_options.monolith_base_payload.pop('warehouseType', DATA_LAKE_WAREHOUSE_TYPE)

        result = self._validate_connection(
            query=onboarding_config.validation_query,
            response_field=onboarding_config.validation_response,
            **onboarding_config.connection_options.monolith_base_payload
        )

        if not onboarding_config.connection_options.validate_only:
            prompt = VALIDATIONS_FAILED_VERBIAGE if result.has_warnings else CONFIRM_CONNECTION_VERBIAGE
            prompt_connection(message=prompt, skip_prompt=onboarding_config.connection_options.auto_yes)
            self._add_connection(temp_path=result.credentials_key, onboarding_config=onboarding_config)
        else:
            click.echo(SKIP_ADD_CONNECTION_VERBIAGE)

    def handle_cert(self, cert_prefix: str, options: Dict) -> None:
        """
        Handles cert payload from either an s3 path or file. Uploading the latter
        Options is updated if successful.
        """
        if options.get('cert_file') is not None:
            bucket_name = self._get_dc_property(prop='PrivateS3BucketArn').split(':')[5]  # get name from arn
            object_name = os.path.join(cert_prefix, os.path.basename(options['cert_file']))
            self._aws_wrapper.upload_file(bucket_name=bucket_name, object_name=object_name,
                                          file_path=options['cert_file'])

            click.echo(f"Uploaded '{options['cert_file']}' to s3://{bucket_name}/{object_name}")
            options['cert_s3'] = object_name

        ssl_options = {}
        if options.get('cert_s3') is not None:
            # reformat to generic options and specify a mechanism
            ssl_options = {'mechanism': S3_CERT_MECHANISM, 'cert': options.pop('cert_s3')}
        options.pop('cert_file', None)

        if options.get('skip_cert_verification') is not None:
            ssl_options['skip_verification'] = options.pop('skip_cert_verification')

        if ssl_options:
            options['ssl_options'] = ssl_options

    def _validate_connection(self, query: str, response_field: str, **kwargs) -> ValidationResult:
        """
        Validate connection before adding using the expected gql response_field (e.g. testPrestoCredentials)
        """
        response = self._request_wrapper.make_request_v2(
            query=query,
            operation=response_field,
            variables=kwargs
        )

        summary = self._build_validation_summary(response.data)
        temp_path = response.data.get('key')
        warnings = response.data.get('warnings')

        if temp_path is not None:
            click.echo(f'{CONNECTION_TEST_SUCCESS_VERBIAGE}{summary}')
            return ValidationResult(
                has_warnings=warnings is not None and len(warnings) > 0,
                credentials_key=temp_path
            )

        complain_and_abort(f'{CONNECTION_TEST_FAILED_VERBIAGE}{summary}')

    def _build_validation_summary(self, response: Dict, table_format: Optional[str] = 'fancy_grid') -> str:
        data = []
        if response.get('validations'):
            for passed in response.get('validations'):
                data.append([
                    passed.get('message'),
                    self._VALIDATION_PASSED,
                    None
                ])

        if response.get('warnings'):
            for failed in response.get('warnings', []):
                data.append([
                    failed.get('message'),
                    self._VALIDATION_FAILED,
                    failed.get('data', {}).get('error')
                ])

        if data:
            return '\n' + tabulate(data, headers=self._VALIDATION_STATUS_HEADERS, tablefmt=table_format)

        # do not provide a summary if no detailed response is available
        return ''

    def _add_connection(self, temp_path: str, onboarding_config: OnboardingConfiguration) -> bool:
        """
        Add connection and setup any associated jobs
        """
        warehouse_type = onboarding_config.warehouse_type or DATA_LAKE_WAREHOUSE_TYPE  # default to data-lake
        connection_request = {'key': temp_path, 'connectionType': onboarding_config.connection_type}
        connection_request.update(self._disambiguate_warehouses(warehouse_type))

        # Set optional properties
        if onboarding_config.warehouse_name:
            connection_request['name'] = onboarding_config.warehouse_name
        if onboarding_config.job_types:
            connection_request['jobTypes'] = onboarding_config.job_types
        if onboarding_config.job_limits:
            connection_request['jobLimits'] = json.dumps(onboarding_config.job_limits)
        if onboarding_config.connection_options and onboarding_config.connection_options.dc_id:
            connection_request['dcId'] = onboarding_config.connection_options.dc_id

        response = self._request_wrapper.make_request_v2(
            query=onboarding_config.connection_query or ADD_CONNECTION_MUTATION,
            operation=onboarding_config.connection_response or EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
            variables=connection_request,
        )
        connection_id = response.data.get('connection', {}).get('uuid')
        if connection_id is not None:
            click.echo(f"{ADD_CONNECTION_SUCCESS_VERBIAGE}"
                       f"{GQL_TO_FRIENDLY_CONNECTION_MAP.get(onboarding_config.connection_type, onboarding_config.connection_type.capitalize())}.")
            return True
        complain_and_abort(ADD_CONNECTION_FAILED_VERBIAGE)

    def _disambiguate_warehouses(self, warehouse_type: str) -> Dict:
        """
        Determine type of connection request to build using the following criteria -
            1) Does an existing warehouse exist and is it a data-lake? If so, attach connection.
            2) Otherwise, it is either the first connection or traditional warehouse (e.g. redshift, snowflake, etc.)

        Multiple lakes are not supported.
        """
        if len(self._user_service.warehouses) > 0 and warehouse_type == DATA_LAKE_WAREHOUSE_TYPE:
            for warehouse in self._user_service.warehouses:
                if normalize_gql(warehouse['connectionType']) == DATA_LAKE_WAREHOUSE_TYPE:
                    return {'dwId': warehouse['uuid']}
        return {'createWarehouseType': warehouse_type}

    def _get_dc_property(self, prop: str) -> Optional[str]:
        """
        Retrieve property from DC stack outputs
        """
        self._dc_outputs = self._dc_outputs or self._aws_wrapper.get_stack_outputs(
            self._user_service.active_collector['stackArn'])  # cache lookup
        for output in self._dc_outputs:
            if output['OutputKey'] == prop:
                return output['OutputValue']

    @staticmethod
    def _build_connection_options(**kwargs) -> ConnectionOptions:
        """
        Create connection options from arguments
        """
        connection_options = ConnectionOptions(**kwargs)

        if connection_options.monolith_connection_payload:
            connection_options.monolith_base_payload['connectionOptions'] = connection_options.monolith_connection_payload
        return connection_options



from typing import Dict, Optional, List

import click
import click_config_file

from montecarlodata import settings
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.discovery.networking import NetworkDiscoveryService
from montecarlodata.tools import add_common_options

# Shared command verbiage
PROFILE_VERBIAGE = 'If not specified, the one in the Monte Carlo CLI profile is used'
RESOURCE_VERBIAGE = 'This can be helpful if the resource and collector are in different accounts'


@click.group(help='Display information about resources.')
def discovery():
    """
    Group for any discovery related subcommands
    """
    pass


@discovery.command(help='List details about EMR clusters in a region.')
@click.option('--aws-profile', help=f'AWS profile. {PROFILE_VERBIAGE}', required=False)
@click.option('--aws-region', help=f'AWS region. {PROFILE_VERBIAGE}', required=False)
@click.option('--only-log-locations', help='Display only unique log locations', is_flag=True, default=False)
@click.option('--created-after', help='Display clusters created after date (e.g. 2017-07-04T00:01:30)', required=False)
@click.option('--state', help='Cluster states', required=False, type=click.Choice(['active', 'terminated', 'failed']),
              multiple=True)
@click.option('--no-grid',
              help='Do not display as grid and print as results are available, useful when the cluster list is large',
              is_flag=True, default=False)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
@click.pass_obj
def list_emr_clusters(ctx: Dict, aws_profile: Optional[str] = None, aws_region: Optional[str] = None,
                      only_log_locations: Optional[bool] = False, created_after: Optional[str] = None,
                      state: Optional[List] = None, no_grid: Optional[bool] = False) -> None:
    CloudResourceService(config=ctx['config'], aws_profile_override=aws_profile, aws_region_override=aws_region) \
        .list_emr_clusters(only_log_locations=only_log_locations, created_after=created_after, states=state,
                           no_grid=no_grid)


@discovery.command(help='Alpha network recommender. Attempts to analyze and makes recommendations on how to connect a '
                        'resource with the Data Collector.')
@click.pass_obj
@click.option('--resource-identifier', required=True,
              help='Identifier for the AWS resource you want to connect the Collector with (e.g. Redshift cluster ID).')
@click.option('--resource-type', required=True, help='Type of AWS resource.',
              type=click.Choice(NetworkDiscoveryService.MCD_NETWORK_REC_RESOURCE_TYPE_MAP.keys()))
@click.option('--resource-aws-region', required=False,
              help='Override the AWS region where the resource is located. Defaults to the region where the collector is hosted.')
@click.option('--resource-aws-profile', required=False,
              help=f'Override the AWS profile use by the CLI for the resource. {RESOURCE_VERBIAGE}.')
@click.option('--collector-aws-profile', required=False,
              help=f'Override the AWS profile use by the CLI for the Collector. {RESOURCE_VERBIAGE}.')
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def network_recommender(ctx, **kwargs):
    NetworkDiscoveryService(config=ctx['config'], aws_wrapper=None).recommend_network_dispatcher(**kwargs)



# Queries related to users

GET_USER_QUERY = """
query getUser {
  getUser {
    firstName
    lastName
    account {
      uuid
      activeCollectionRegions
      warehouses {
        uuid
        name
        connectionType
        connections {
          uuid
          type
          createdOn
          jobTypes
        }
        dataCollector {
          uuid
          customerAwsRegion
        }
      }
      bi {
        uuid
        connections {
          uuid
          type
          createdOn
          jobTypes
        }
        dataCollector {
          uuid,
          customerAwsRegion
        }
      }
      tableauAccounts {
        uuid
        dataCollector {
          uuid,
          customerAwsRegion
        }
      }
      dataCollectors {
        uuid
        stackArn
        active
        customerAwsAccountId
        templateVersion
        lastUpdated
      }
    }
  }
}
"""



import click
import click_config_file

from montecarlodata import settings
from montecarlodata.iac.mc_config_service import MonteCarloConfigService


@click.group(help='Manage monitors.')
def monitors():
    """
    Group for any collector related subcommands
    """
    pass

@monitors.command(help='Compile monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.pass_obj
def compile(ctx, project_dir):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).compile()

@monitors.command(help='Compile and apply monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.option('--namespace', required=True, help='Namespace of monitors configuration.')
@click.option('--dry-run', required=False, default=False, show_default=True, is_flag=True, help='Dry run (just shows planned changes but doesn\'t apply them.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def apply(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).apply(namespace, dry_run=dry_run)

@monitors.command(help='Delete monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.option('--namespace', required=True, help='Namespace of monitors configuration.')
@click.option('--dry-run', required=False, default=False, show_default=True, is_flag=True, help='Dry run (just shows planned changes but doesn\'t apply them.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).delete(namespace, dry_run=dry_run)


# Queries related to onboarding

TEST_PRESTO_CRED_MUTATION = """
mutation testPrestoCredentials($catalog:String, $host:String, $httpScheme:String, $password:String, $port:Int, $schema:String, $sslOptions:SslInputOptions, $user:String, $connectionOptions:ConnectionTestOptions) {
  testPrestoCredentials(catalog:$catalog, host:$host, httpScheme:$httpScheme, password:$password, port:$port, schema:$schema, sslOptions: $sslOptions, user:$user, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_S3_CRED_MUTATION = """
mutation tests3Credentials($bucket:String, $prefix:String, $assumableRole:String, $externalId:String, $connectionType:String, $connectionOptions:ConnectionTestOptions) {
  testS3Credentials(bucket:$bucket, prefix:$prefix, assumableRole:$assumableRole, externalId:$externalId, connectionType:$connectionType, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_DATABASE_CRED_MUTATION = """
mutation testDatabaseCredentials($assumableRole:String, $connectionType:String, $dbName:String, $externalId:String, $host:String, $password:String, $port:Int, $sslOptions:SslInputOptions, $user:String, $connectionOptions:ConnectionTestOptions) {
  testDatabaseCredentials(assumableRole:$assumableRole, connectionType:$connectionType, dbName:$dbName, externalId:$externalId, host:$host, password:$password, port:$port, sslOptions:$sslOptions, user:$user, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_HIVE_SQL_CRED_MUTATION = """
mutation testHiveCredentials($database:String, $host:String, $port:Int, $username:String, $authMode:String, $connectionOptions:ConnectionTestOptions) {
  testHiveCredentials(database:$database, host:$host, port:$port, username:$username, authMode:$authMode, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_GLUE_CRED_MUTATION = """
mutation testGlueCredentials($assumableRole:String, $externalId:String, $awsRegion:String, $connectionOptions:ConnectionTestOptions) {
  testGlueCredentials(assumableRole:$assumableRole, externalId:$externalId, awsRegion:$awsRegion, connectionOptions:$connectionOptions) {
    success
    key,
    validations {
      type,
      message,
      data {
        error
      }
    },
    warnings {
      type,
      message,
      data {
        error
      }
    }
  }
}
"""

TEST_ATHENA_CRED_MUTATION = """
mutation testAthenaCredentials($assumableRole:String, $catalog:String, $externalId:String, $workgroup:String, $awsRegion:String, $connectionOptions:ConnectionTestOptions) {
  testAthenaCredentials(assumableRole:$assumableRole, catalog:$catalog, externalId:$externalId, workgroup:$workgroup, awsRegion:$awsRegion, connectionOptions:$connectionOptions) {
    success
    key,
    validations {
      type,
      message,
      data {
        error
      }
    },
    warnings {
      type,
      message,
      data  {
        error
      }
    }
  }
}
"""

TEST_SPARK_BINARY_MODE_CRED_MUTATION = """
mutation testSparkCredentials($database:String!, $host:String!, $port:Int!, $username:String!, $password:String!, $connectionOptions:ConnectionTestOptions) {
  testSparkCredentials(binaryMode: {database:$database, host:$host, port:$port, username:$username, password:$password}, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_SPARK_HTTP_MODE_CRED_MUTATION = """
mutation testSparkCredentials($url:String!, $username:String!, $password:String!, $connectionOptions:ConnectionTestOptions) {
  testSparkCredentials(httpMode: {url:$url, username:$username, password:$password}, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_SPARK_DATABRICKS_CRED_MUTATION = """
mutation testSparkCredentials($workspaceUrl:String!, $workspaceId:String!, $clusterId:String!, $token:String!, $connectionOptions:ConnectionTestOptions) {
  testSparkCredentials(databricks: {workspaceUrl:$workspaceUrl, workspaceId:$workspaceId, clusterId:$clusterId, token:$token}, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_SNOWFLAKE_CRED_MUTATION = """
mutation testSnowflakeCredentials($account:String, $password:String, $user:String, $warehouse:String, $connectionOptions:ConnectionTestOptions) {
  testSnowflakeCredentials(account:$account, password:$password, user:$user, warehouse:$warehouse, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_BQ_CRED_MUTATION = """
mutation testBqCredentials($serviceJson:String, $connectionOptions:ConnectionTestOptions) {
  testBqCredentials(serviceJson:$serviceJson, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TEST_SELF_HOSTED_CRED_MUTATION = """
mutation testSelfHostedCredentials($connectionType:String, $selfHostingMechanism:String, $selfHostingKey:String, $assumableRole:String, $externalId:String, $region:String, $connectionOptions:ConnectionTestOptions) {
  testSelfHostedCredentials(connectionType:$connectionType, selfHostingMechanism:$selfHostingMechanism, selfHostingKey:$selfHostingKey, assumableRole:$assumableRole, externalId:$externalId, region:$region, connectionOptions:$connectionOptions) {
    key
  }
}
"""

TOGGLE_EVENT_MUTATION = """
mutation toggleEventConfig($dwId:UUID!, $enable:Boolean!, $connectionType:String!, $eventType:DataCollectorEventTypes!, $assumableRole:String, $externalId:String, $formatType:String, $sourceFormat:String, $mapping:JSONString) {
  toggleEventConfig(dwId:$dwId, enable:$enable, connectionType:$connectionType, eventType: $eventType, assumableRole:$assumableRole, externalId:$externalId, formatType:$formatType, sourceFormat:$sourceFormat, mapping:$mapping){
    success
  }
}
"""

ADD_CONNECTION_MUTATION = """
mutation addConnection($connectionType:String!, $createWarehouseType:String, $dwId:UUID, $jobTypes:[String], $key:String!, $jobLimits:JSONString, $name:String, $dcId:UUID) {
  addConnection(connectionType:$connectionType, createWarehouseType:$createWarehouseType, dwId:$dwId, jobTypes:$jobTypes, key:$key, jobLimits:$jobLimits, name:$name, dcId: $dcId){
    connection {
      uuid
    }
  }
}
"""

UPDATE_CREDENTIALS_MUTATION = """
mutation updateCredentials($connectionId: UUID!, $changes:JSONString!, $shouldValidate:Boolean, $shouldReplace:Boolean) {
  updateCredentials(connectionId:$connectionId, changes:$changes, shouldValidate:$shouldValidate, shouldReplace:$shouldReplace){
    success
  }
}
"""

REMOVE_CONNECTION_MUTATION = """
mutation removeConnection($connectionId: UUID!) {
  removeConnection(connectionId:$connectionId) {
    success
  }
}
"""

TEST_EXISTING_CONNECTION_QUERY = """
query testExistingConnection($connectionId:UUID) {
  testExistingConnection(connectionId:$connectionId) {
    success
    validations {
      type
      message
      data {
        database
        table
        error
      }
    }
    warnings {
      type
      message
      data {
        database
        table
        error
      }
    }
  }
}
"""

TEST_TABLEAU_CRED_MUTATION = """
mutation testTableauCredentials($connectionOptions:ConnectionTestOptions, $password:String, $serverName:String!, $siteName:String!, $tokenName:String, $tokenValue:String, $username:String, $verifySsl:Boolean) {
  testTableauCredentials(connectionOptions:$connectionOptions, password:$password, serverName:$serverName, siteName:$siteName, tokenName:$tokenName, tokenValue:$tokenValue, username:$username, verifySsl:$verifySsl) {
    success
  }
}

"""

ADD_TABLEAU_CONNECTION_MUTATION = """
mutation addTableauAccount($dcId:UUID, $password:String, $serverName:String!, $siteName:String!, $tokenName:String, $tokenValue:String, $username:String, $verifySsl:Boolean) {
  addTableauAccount(dcId:$dcId, password:$password, serverName:$serverName, siteName:$siteName, tokenName:$tokenName, tokenValue:$tokenValue, username:$username, verifySsl:$verifySsl) {
    tableauAccount {
      uuid
    }
  }
}
"""

TEST_LOOKER_METADATA_CRED_MUTATION = """
mutation testLookerCredentials($baseUrl:String, $clientId:String, $clientSecret:String, $connectionOptions:ConnectionTestOptions, $verifySsl:Boolean) {
  testLookerCredentials(baseUrl:$baseUrl, clientId:$clientId, clientSecret:$clientSecret, connectionOptions:$connectionOptions, verifySsl:$verifySsl) {
    key
    success
  }
}
"""

TEST_LOOKER_GIT_SSH_CRED_MUTATION = """
mutation testLookerGitSshCredentials($connectionOptions:ConnectionTestOptions, $repoUrl:String, $sshKey:String) {
  testLookerGitSshCredentials(connectionOptions:$connectionOptions, repoUrl:$repoUrl, sshKey:$sshKey) {
    key
    success
  }
}
"""

TEST_LOOKER_GIT_CLONE_CRED_MUTATION = """
mutation testLookerGitCloneCredentials($connectionOptions:ConnectionTestOptions, $repoUrl:String!, $sshKey:String, $username:String, $token:String) {
  testLookerGitCloneCredentials(connectionOptions:$connectionOptions, repoUrl:$repoUrl, sshKey:$sshKey, username:$username, token:$token) {
    key
    success
  }
}
"""

ADD_BI_CONNECTION_MUTATION = """
mutation addBiConnection($connectionType:String!, $key:String!, $dcId:UUID) {
  addBiConnection(connectionType:$connectionType, key:$key, dcId:$dcId){
    connection {
      uuid
    }
  }
}
"""



import distutils.text_file
from pathlib import Path
from typing import List, Optional

from setuptools import setup, find_packages

__VERSION__ = '0.15.1'


def parse_requirements(file_name: Optional[str] = 'requirements.txt') -> List[str]:
    return distutils.text_file.TextFile(filename=str(Path(__file__).with_name(file_name))).readlines()


def get_long_description():
    with open('README.md', 'r') as fh:
        return fh.read()


setup(
    name='montecarlodata',
    version=__VERSION__,
    license='Apache Software License (Apache 2.0)',
    description='Monte Carlo\'s CLI',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Monte Carlo Data, Inc',
    author_email='info@montecarlodata.com',
    url='https://www.montecarlodata.com/',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=parse_requirements(),
    entry_points='''
        [console_scripts]
        montecarlo=montecarlodata.cli:entry_point
    ''',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    python_requires='>=3.7'
)



import click
from pycarlo.features.dbt import DbtImporter, DbtCloudImporter

from montecarlodata.common.common import create_mc_client


@click.group(help='Import data.', name='import')
def import_subcommand():
    """
    Group for any import related subcommands
    """
    pass


@import_subcommand.command(help='Import DBT manifest.')
@click.argument('MANIFEST_FILE', required=True, type=click.Path(exists=True))
@click.option('--project-name', required=False, type=click.STRING,
              help='Name that uniquely identifies dbt project.')
@click.option('--batch-size', required=False, default=10, type=click.INT,
              help='Number of DBT manifest nodes to send in each batch.'
                   'Use smaller number if requests are timing out.'
                   'Use larger number for higher throughput.')
@click.option('--default-resource', required=False, type=click.STRING,
              help='The warehouse friendly name or UUID where dbt objects will be associated with.')
@click.pass_obj
def dbt_manifest(ctx, manifest_file, project_name, batch_size, default_resource):
    DbtImporter(mc_client=create_mc_client(ctx), print_func=click.echo).import_dbt_manifest(
        dbt_manifest=manifest_file,
        project_name=project_name,
        batch_size=batch_size,
        default_resource=default_resource
    )


@import_subcommand.command(help='Import DBT run results.')
@click.argument('RUN_RESULTS_FILE', required=True, type=click.Path(exists=True))
@click.option('--project-name', required=False, type=click.STRING,
              help='Name that uniquely identifies dbt project.')
@click.pass_obj
def dbt_run_results(ctx, run_results_file, project_name):
    DbtImporter(mc_client=create_mc_client(ctx), print_func=click.echo).import_run_results(
        dbt_run_results=run_results_file,
        project_name=project_name
    )


@import_subcommand.command(help='Import manifest and run results from DBT cloud. This command is experimental.')
@click.option('--project-id', required=False, type=click.STRING,
              help='dbt cloud project ID to import. If not specified, all projects will be imported.')
@click.option('--job-id', required=False, type=click.STRING,
              help='dbt cloud job ID to import. If not specified, all jobs will be imported.')
@click.option('--manifest-only', required=False, type=click.BOOL, is_flag=True,
              help='If used, will only import manifest (not run results)')
@click.option('--default-resource', required=False, type=click.STRING,
              help='The warehouse friendly name or UUID where dbt objects will be associated with.')
@click.pass_obj
def dbt_cloud(ctx, project_id, job_id, manifest_only, default_resource):
    DbtCloudImporter(client=create_mc_client(ctx), print_func=click.echo).import_dbt_cloud(
        project_id=project_id,
        job_id=job_id,
        manifest_only=manifest_only,
        default_resource=default_resource
    )




import copy
from typing import Optional, Dict, List
from urllib import parse

import click
from tabulate import tabulate

from montecarlodata.collector.fields import EXPECTED_GENERATE_TEMPLATE_GQL_RESPONSE_FIELD, DEFAULT_COLLECTION_REGION, \
    EXPECTED_ADD_DC_RESPONSE_FIELD, ADD_DC_PROMPT_VERBIAGE, ADD_DC_REGION_PROMPT_VERBIAGE
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort, prompt_connection
from montecarlodata.queries.collector import GENERATE_COLLECTOR_TEMPLATE, ADD_COLLECTOR_RECORD
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class CollectorManagementService:
    _DEFAULT_COLLECTOR_VAL = '-'
    _LIST_COLLECTOR_HEADERS = ['AWS Stack ARN', 'ID', 'Version', 'Last updated', 'Active']

    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None, aws_profile_override: Optional[str] = None,
                 aws_region_override: Optional[str] = None, user_service: Optional[UserService] = None):
        self._abort_on_error = True
        self._collector_region = aws_region_override or config.aws_region or DEFAULT_COLLECTION_REGION

        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=aws_profile_override or config.aws_profile,
                                                            region_name=self._collector_region)
        self._user_service = user_service or UserService(request_wrapper=self._request_wrapper, config=config)

    @manage_errors
    def echo_template(self, **kwargs) -> None:
        """
        Echos the most recent template for this account
        """
        click.echo(self._get_template_url_from_launch(self._get_template_launch_url(**kwargs)))

    @manage_errors
    def echo_collectors(self, headers: Optional[str] = 'firstrow', table_format: Optional[str] = 'fancy_grid',
                        active_only: Optional[bool] = False) -> None:
        """
        Echo collectors in a fancy grid.
        """
        table = [self._LIST_COLLECTOR_HEADERS]
        for collector in self._user_service.collectors or []:
            if active_only and not collector.get('active'):
                continue
            table.append(
                [
                    collector.get('stackArn') if collector.get('stackArn') else self._DEFAULT_COLLECTOR_VAL,
                    collector['uuid'],
                    collector.get('templateVersion') if collector.get('templateVersion') else self._DEFAULT_COLLECTOR_VAL,
                    collector.get('lastUpdated') if collector.get('lastUpdated') else self._DEFAULT_COLLECTOR_VAL,
                    collector.get('active')
                ]
            )
        click.echo(tabulate(table, headers=headers, tablefmt=table_format))

    @manage_errors
    def add_collector(self, no_prompt: Optional[bool] = False) -> None:
        """
        Add a collector to the account. Prompts to launch browser.
        """
        response = self._request_wrapper.make_request_v2(
            query=ADD_COLLECTOR_RECORD,
            operation=EXPECTED_ADD_DC_RESPONSE_FIELD,
        )
        dc_id = response.data.dc.uuid
        click.echo(f'Created collector record with ID \'{dc_id}\'')
        if no_prompt:
            return

        prompt_connection(message=ADD_DC_PROMPT_VERBIAGE)  # exits on No
        collection_region = click.prompt(
            ADD_DC_REGION_PROMPT_VERBIAGE,
            show_choices=True,
            default=DEFAULT_COLLECTION_REGION,
            show_default=True,
            type=click.Choice(self._user_service.active_collection_regions)
        )
        self.launch_quick_create_link(dry=False, dc_id=dc_id, collection_region=collection_region)

    @manage_errors
    def launch_quick_create_link(self, dry: Optional[bool] = True, **kwargs) -> None:
        """
        Open browser with a quick create link for deploying a data collector

        https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cfn-console-create-stacks-quick-create-links.html
        """
        launch_url = self._get_template_launch_url(**kwargs)
        if dry:
            click.echo(launch_url)
            return
        click.launch(launch_url)

    @manage_errors
    def upgrade_template(self, new_params: Optional[Dict] = None, **kwargs) -> None:
        """
        Upgrades the DC attached to this account
        """
        collector_details = self._generate_collector_template(**kwargs)
        try:
            stack_arn = collector_details['stackArn']
            is_active = collector_details['active']
            gateway_id = collector_details['apiGatewayId']
            template_url = self._get_template_url_from_launch(collector_details['templateLaunchUrl'])

            if not is_active:
                complain_and_abort('Cannot upgrade an inactive collector. Please contact Monte Carlo')
        except KeyError as err:
            complain_and_abort(f'Missing expected property ({err}). The collector may not have been deployed before')
        else:
            click.echo(f'Updating \'{stack_arn}\'')
            self._upgrade(stack_arn=stack_arn, template_url=template_url, gateway_id=gateway_id, new_params=new_params)
            click.echo('Upgrade complete! Have a nice day!')

    @manage_errors
    def deploy_template(self, stack_name: str, termination_protection: bool, new_params: Optional[Dict] = None,
                        **kwargs) -> None:
        """
        Deploys a template for this account
        """
        click.echo(f'Deploying \'{stack_name}\' in \'{self._collector_region}\'')
        collector_details = self._generate_collector_template(**kwargs)
        is_active = collector_details['active']
        if is_active:
            complain_and_abort('Cannot deploy with an already active collector.')

        template_url = self._get_template_url_from_launch(collector_details['templateLaunchUrl'])
        self._deploy(stack_name=stack_name, template_url=template_url, termination_protection=termination_protection,
                     new_params=new_params)

    def _upgrade(self, stack_arn: str, template_url: str, gateway_id: str, new_params: Dict) -> None:
        """
        Upgrade the collector stack and deploy the gateway
        """
        parameters = self._build_param_list(self._aws_wrapper.get_stack_parameters(stack_id=stack_arn), new_params)
        if self._aws_wrapper.upgrade_stack(stack_id=stack_arn, template_link=template_url, parameters=parameters):
            self._aws_wrapper.deploy_gateway(gateway_id=gateway_id)
        else:
            complain_and_abort('Failed to upgrade. Please review CF events for details')

    def _deploy(self, stack_name: str, template_url: str, termination_protection: bool,
                new_params: Optional[Dict] = None) -> None:
        """
        Deploy (create) a collector stack for the first time
        """
        parameters = self._build_param_list(existing_params=[], new_params=new_params)
        if not self._aws_wrapper.create_stack(stack_name=stack_name, template_link=template_url,
                                              termination_protection=termination_protection,
                                              parameters=parameters):
            complain_and_abort('Failed to deploy. Please review CF events for details')

    def _generate_collector_template(self, **kwargs) -> Optional[Dict]:
        """
        Generate the latest template and returns any associated dc properties (i.e. from the DataCollectorModel)
        """
        options = {'region': self._collector_region}
        if kwargs.get('collection_region'):
            options['region'] = kwargs['collection_region']  # overwrite client

        if kwargs.get('dc_id'):
            options['dc_id'] = kwargs['dc_id']

        response = self._request_wrapper.make_request_v2(
            query=GENERATE_COLLECTOR_TEMPLATE,
            operation=EXPECTED_GENERATE_TEMPLATE_GQL_RESPONSE_FIELD,
            variables=options
        )
        return response.data.get('dc')

    def _get_template_launch_url(self, **kwargs) -> str:
        """
        Get the template launch url
        """
        return self._generate_collector_template(**kwargs).get('templateLaunchUrl')

    @staticmethod
    def _get_template_url_from_launch(launch_url: str) -> str:
        """
        Extract the template url from expected launch url structure
        """
        return parse.parse_qs(parse.urlparse(launch_url).fragment).get('templateURL', [])[0]

    @staticmethod
    def _build_param_list(existing_params: List[Dict], new_params: Optional[Dict] = None) -> List[Dict]:
        """
        Get a list of parameters by replacing (or extending) any new params into the existing stack params
        """
        existing_params = copy.deepcopy(existing_params)
        new_params = copy.deepcopy(new_params)

        for param in existing_params or []:
            if new_params and new_params.get(param['ParameterKey']) is not None:
                param['ParameterValue'] = new_params[param['ParameterKey']]
                param['UsePreviousValue'] = False
                del new_params[param['ParameterKey']]
            else:
                del param['ParameterValue']
                param['UsePreviousValue'] = True
        if new_params:
            # Handle any completely new params (i.e. those that were not in the existing struct)
            # Can be found for params added between versions
            existing_params.extend([{'ParameterKey': k, 'ParameterValue': v} for k, v in new_params.items()])

        return existing_params
